# vim: set fileencoding=utf8 :
from __future__ import with_statement
from __future__ import unicode_literals
from decimal import Decimal, Context
import uuid
import datetime
import logging
import random
import string
import pytest
import six
import os
from six import StringIO, BytesIO
import pytds
import pytds.extensions
import pytds.login
import settings

from fixtures import *


logger = logging.getLogger(__name__)
LIVE_TEST = getattr(settings, 'LIVE_TEST', True)
pytds.tds.logging_enabled = True


def test_integrity_error(cursor):
    cursor.execute('create table testtable_pk(pk int primary key)')
    cursor.execute('insert into testtable_pk values (1)')
    with pytest.raises(pytds.IntegrityError):
        cursor.execute('insert into testtable_pk values (1)')


def test_rollback_timeout_recovery(separate_db_connection):
    conn = separate_db_connection
    conn.autocommit = False
    with conn.cursor() as cur:
        cur.execute('''
        create table testtable_rollback (field int)
        ''')
        sql = 'insert into testtable_rollback values ' + ','.join(['(1)'] * 1000)
        for i in range(10):
            cur.execute(sql)

    conn._conn.sock.settimeout(0.00001)
    try:
        conn.rollback()
    except:
        pass

    conn._conn.sock.settimeout(10)
    cur = conn.cursor()
    cur.execute('select 1')
    cur.fetchall()


def test_commit_timeout_recovery(separate_db_connection):
    conn = separate_db_connection
    conn.autocommit = False
    with conn.cursor() as cur:
        try:
            cur.execute('drop table testtable_commit_rec')
        except:
            pass
        cur.execute('''
        create table testtable_commit_rec (field int)
        ''')
        sql = 'insert into testtable_commit_rec values ' + ','.join(['(1)'] * 1000)
        for i in range(10):
            cur.execute(sql)

    conn._conn.sock.settimeout(0.00001)
    try:
        conn.commit()
    except:
        pass

    conn._conn.sock.settimeout(10)
    cur = conn.cursor()
    cur.execute('select 1')
    cur.fetchall()


def test_autocommit(separate_db_connection):
    conn = separate_db_connection
    assert not conn.autocommit
    with conn.cursor() as cur:
        try:
            cur.execute('drop table test_autocommit')
        except:
            pass
        cur.execute('create table test_autocommit(field int)')
        conn.commit()
        assert 1 == conn._trancount()
        cur.execute('insert into test_autocommit(field) values(1)')
        assert 1 == conn._trancount()
        cur.execute('select field from test_autocommit')
        row = cur.fetchone()
        conn.rollback()
        cur.execute('select field from test_autocommit')
        row = cur.fetchone()
        assert not row

        conn.autocommit = True
        # commit in autocommit mode should be a no-op
        conn.commit()
        # rollback in autocommit mode should be a no-op
        conn.rollback()
        cur.execute('insert into test_autocommit(field) values(1)')
        assert 0 == conn._trancount()


def test_bulk_insert(cursor):
    cur = cursor
    f = StringIO("42\tfoo\n74\tbar\n")
    cur.copy_to(f, 'bulk_insert_table', schema='myschema', columns=('num', 'data'))
    cur.execute('select num, data from myschema.bulk_insert_table')
    assert [(42, 'foo'), (74, 'bar')] == cur.fetchall()


def test_bug2(cursor):
    cur = cursor
    cur.execute('''
    create procedure testproc_bug2 (@param int)
    as
    begin
        set transaction isolation level read uncommitted -- that will produce very empty result (even no rowcount)
        select @param
        return @param + 1
    end
    ''')
    val = 45
    cur.execute('exec testproc_bug2 @param = 45')
    assert cur.fetchall() == [(val,)]
    assert val + 1 == cur.get_proc_return_status()


def test_stored_proc(cursor):
    cur = cursor
    val = 45
    #params = {'@param': val, '@outparam': output(None), '@add': 1}
    values = cur.callproc('testproc', (val, pytds.default, pytds.output(value=1)))
    #self.assertEqual(cur.fetchall(), [(val,)])
    assert val + 2 == values[2]
    assert val + 2 == cur.get_proc_return_status()


def test_table_selects(db_connection):
    cur = db_connection.cursor()
    cur.execute(u'''
    create table #testtable (id int, _text text, _xml xml, vcm varchar(max), vc varchar(10))
    ''')
    cur.execute(u'''
    insert into #testtable (id, _text, _xml, vcm, vc) values (1, 'text', '<root/>', '', NULL)
    ''')
    cur.execute('select id from #testtable order by id')
    assert [(1,)] == cur.fetchall()

    cur = db_connection.cursor()
    cur.execute('select _text from #testtable order by id')
    assert [(u'text',)] == cur.fetchall()

    cur = db_connection.cursor()
    cur.execute('select _xml from #testtable order by id')
    assert [('<root/>',)] == cur.fetchall()

    cur = db_connection.cursor()
    cur.execute('select id, _text, _xml, vcm, vc from #testtable order by id')
    assert (1, 'text', '<root/>', '', None) == cur.fetchone()

    cur = db_connection.cursor()
    cur.execute('select vc from #testtable order by id')
    assert [(None,)] == cur.fetchall()

    cur = db_connection.cursor()
    cur.execute('insert into #testtable (_xml) values (%s)', ('<some/>',))

    cur = db_connection.cursor()
    cur.execute(u'drop table #testtable')


def test_decimals(cursor):
    cur = cursor
    assert Decimal(12) == cur.execute_scalar('select cast(12 as decimal) as fieldname')
    assert Decimal(-12) == cur.execute_scalar('select cast(-12 as decimal) as fieldname')
    assert Decimal('123456.12345') == cur.execute_scalar("select cast('123456.12345'as decimal(20,5)) as fieldname")
    assert Decimal('-123456.12345') == cur.execute_scalar("select cast('-123456.12345'as decimal(20,5)) as fieldname")


def test_bulk_insert_with_special_chars_no_columns(cursor):
    cur = cursor
    cur.execute('create table [test]] table](num int not null, data varchar(100))')
    f = StringIO("42\tfoo\n74\tbar\n")
    cur.copy_to(f, 'test] table')
    cur.execute('select num, data from [test]] table]')
    assert cur.fetchall() == [(42, 'foo'), (74, 'bar')]


def test_bulk_insert_with_special_chars(cursor):
    cur = cursor
    cur.execute('create table [test]] table](num int, data varchar(100))')
    f = StringIO("42\tfoo\n74\tbar\n")
    cur.copy_to(f, 'test] table', columns=('num', 'data'))
    cur.execute('select num, data from [test]] table]')
    assert cur.fetchall() == [(42, 'foo'), (74, 'bar')]


def test_bulk_insert_with_keyword_column_name(cursor):
    cur = cursor
    cur.execute('create table test_table(num int, [User] varchar(100))')
    f = StringIO("42\tfoo\n74\tbar\n")
    cur.copy_to(f, 'test_table')
    cur.execute('select num, [User] from test_table')
    assert cur.fetchall() == [(42, 'foo'), (74, 'bar')]


def test_bulk_insert_with_direct_data(cursor):
    cur = cursor
    cur.execute('create table test_table(num int, data nvarchar(max))')

    data = [
        [42, 'foo'],
        [57, ''],
        [66, None],
        [74, 'bar']
    ]

    column_types = [
        pytds.tds_base.Column('num', pytds.tds_types.IntType(), pytds.tds_base.Column.fNullable),
        pytds.tds_base.Column('data', pytds.tds_types.NVarCharMaxType(), pytds.tds_base.Column.fNullable)
    ]

    cur.copy_to(data=data, table_or_view='test_table', columns=column_types)
    cur.execute('select num, data from test_table')
    assert cur.fetchall() == [(42, 'foo'), (57, ''), (66, None), (74, 'bar')]


def test_table_valued_type_autodetect(separate_db_connection):
    def rows_gen():
        yield (1, 'test1')
        yield (2, 'test2')

    with separate_db_connection.cursor() as cur:
        cur.execute('CREATE TYPE dbo.CategoryTableType AS TABLE ( CategoryID int, CategoryName nvarchar(50) )')
        separate_db_connection.commit()

        tvp = pytds.TableValuedParam(type_name='dbo.CategoryTableType', rows=rows_gen())
        cur.execute('SELECT * FROM %s', (tvp,))
        assert cur.fetchall() == [(1, 'test1'), (2, 'test2')]

        cur.execute('DROP TYPE dbo.CategoryTableType')
        separate_db_connection.commit()


def test_table_valued_type_explicit(separate_db_connection):
    def rows_gen():
        yield (1, 'test1')
        yield (2, 'test2')

    with separate_db_connection.cursor() as cur:
        cur.execute('CREATE TYPE dbo.CategoryTableType AS TABLE ( CategoryID int, CategoryName nvarchar(50) )')
        separate_db_connection.commit()

        tvp = pytds.TableValuedParam(
            type_name='dbo.CategoryTableType',
            columns=(pytds.Column(type=pytds.tds_types.IntType()), pytds.Column(type=pytds.tds_types.NVarCharType(size=30))),
            rows=rows_gen())
        cur.execute('SELECT * FROM %s', (tvp,))
        assert cur.fetchall() == [(1, 'test1'), (2, 'test2')]

        cur.execute('DROP TYPE dbo.CategoryTableType')
        separate_db_connection.commit()


def test_cursor_env(separate_db_connection):
    with separate_db_connection.cursor() as cursor:
        cursor.execute('use master')
        assert cursor.execute_scalar('select DB_NAME()') == 'master'


def test_minimal(cursor):
    cursor.execute('select 1')
    assert [(1,)] == cursor.fetchall()


def test_empty_query(cursor):
    cursor.execute('')
    assert cursor.description is None


@pytest.mark.parametrize(
    'typ,values',
    [
        (pytds.tds_types.BitType(), [True, False]),
        (pytds.tds_types.IntType(), [2 ** 31 - 1, None]),
        (pytds.tds_types.IntType(), [-2 ** 31, None]),
        (pytds.tds_types.SmallIntType(), [-2 ** 15, 2 ** 15 - 1]),
        (pytds.tds_types.TinyIntType(), [255, 0]),
        (pytds.tds_types.BigIntType(), [2 ** 63 - 1, -2 ** 63]),
        (pytds.tds_types.IntType(), [None, 2 ** 31 - 1]),
        (pytds.tds_types.IntType(), [None, -2 ** 31]),
        (pytds.tds_types.RealType(), [0.25, None]),
        (pytds.tds_types.FloatType(), [0.25, None]),
        (pytds.tds_types.VarCharType(size=10), [u'', u'testtest12', None, u'foo']),
        (pytds.tds_types.VarCharType(size=4000), [u'x' * 4000, u'foo']),
        (pytds.tds_types.VarCharMaxType(), [u'x' * 10000, u'foo', u'', u'testtest', None, u'bar']),
        (pytds.tds_types.NVarCharType(size=10), [u'', u'testtest12', None, u'foo']),
        (pytds.tds_types.NVarCharType(size=4000), [u'x' * 4000, u'foo']),
        (pytds.tds_types.NVarCharMaxType(), [u'x' * 10000, u'foo', u'', u'testtest', None, u'bar']),
        (pytds.tds_types.VarBinaryType(size=10), [b'testtest12', b'', None]),
        (pytds.tds_types.VarBinaryType(size=8000), [b'x' * 8000, b'']),
        (pytds.tds_types.SmallDateTimeType(), [datetime.datetime(1900, 1, 1, 0, 0, 0), None, datetime.datetime(2079, 6, 6, 23, 59, 0)]),
        (pytds.tds_types.DateTimeType(), [datetime.datetime(1753, 1, 1, 0, 0, 0), None, datetime.datetime(9999, 12, 31, 23, 59, 59, 990000)]),
        (pytds.tds_types.DateType(), [datetime.date(1, 1, 1), None, datetime.date(9999, 12, 31)]),
        (pytds.tds_types.TimeType(precision=0), [datetime.time(0, 0, 0), None]),
        (pytds.tds_types.TimeType(precision=6), [datetime.time(23, 59, 59, 999999), None]),
        (pytds.tds_types.TimeType(precision=0), [None]),
        (pytds.tds_types.DateTime2Type(precision=0), [datetime.datetime(1, 1, 1, 0, 0, 0), None]),
        (pytds.tds_types.DateTime2Type(precision=6), [datetime.datetime(9999, 12, 31, 23, 59, 59, 999999), None]),
        (pytds.tds_types.DateTime2Type(precision=0), [None]),
        (pytds.tds_types.DateTimeOffsetType(precision=6), [datetime.datetime(9999, 12, 31, 23, 59, 59, 999999, pytds.tz.utc), None]),
        (pytds.tds_types.DateTimeOffsetType(precision=6), [datetime.datetime(9999, 12, 31, 23, 59, 59, 999999, pytds.tz.FixedOffsetTimezone(14)), None]),
        (pytds.tds_types.DateTimeOffsetType(precision=0), [datetime.datetime(1, 1, 1, 0, 0, 0, tzinfo=pytds.tz.FixedOffsetTimezone(-14))]),
        (pytds.tds_types.DateTimeOffsetType(precision=0), [datetime.datetime(1, 1, 1, 0, 14, 0, tzinfo=pytds.tz.FixedOffsetTimezone(14))]),
        (pytds.tds_types.DateTimeOffsetType(precision=6), [None]),
        (pytds.tds_types.DecimalType(scale=6, precision=38), [Decimal('123.456789'), None]),
        (pytds.tds_types.SmallMoneyType(), [Decimal('214748.3647'), None, Decimal('-214748.3648')]),
        (pytds.tds_types.MoneyType(), [Decimal('922337203685477.5807'), None, Decimal('-922337203685477.5808')]),
        (pytds.tds_types.SmallMoneyType(), [Decimal('214748.3647')]),
        (pytds.tds_types.MoneyType(), [Decimal('922337203685477.5807')]),
        (pytds.tds_types.MoneyType(), [None]),
        (pytds.tds_types.UniqueIdentifierType(), [None, uuid.uuid4()]),
        (pytds.tds_types.VariantType(), [None]),
        #(pytds.tds_types.VariantType(), [100]),
        #(pytds.tds_types.ImageType(), [None]),
        (pytds.tds_types.VarBinaryMaxType(), [None]),
        #(pytds.tds_types.NTextType(), [None]),
        #(pytds.tds_types.TextType(), [None]),
        #(pytds.tds_types.ImageType(), [b'']),
        #(self.conn._conn.type_factory.long_binary_type(), [b'testtest12']),
        #(self.conn._conn.type_factory.long_string_type(), [None]),
        #(self.conn._conn.type_factory.long_varchar_type(), [None]),
        #(self.conn._conn.type_factory.long_string_type(), ['test']),
        #(pytds.tds_types.ImageType(), [None]),
        #(pytds.tds_types.ImageType(), [None]),
        #(pytds.tds_types.ImageType(), [b'test']),
])
def test_bulk_insert_type(cursor, typ, values):
    cur = cursor
    cur.execute('create table bulk_insert_table_ll(c1 {0})'.format(typ.get_declaration()))
    cur._session.submit_plain_query('insert bulk bulk_insert_table_ll (c1 {0})'.format(typ.get_declaration()))
    cur._session.process_simple_request()
    col1 = pytds.Column(name='c1', type=typ, flags=pytds.Column.fNullable)
    metadata = [col1]
    cur._session.submit_bulk(metadata, [[value] for value in values])
    cur._session.process_simple_request()
    cur.execute('select c1 from bulk_insert_table_ll')
    assert cur.fetchall() == [(value,) for value in values]
    assert cur.fetchone() is None


def test_streaming(cursor):
    val = 'x' * 10000
    # test nvarchar(max)
    cursor.execute("select N'{}', 1".format(val))
    with pytest.raises(ValueError):
        cursor.set_stream(1, StringIO())
    with pytest.raises(ValueError):
        cursor.set_stream(2, StringIO())
    with pytest.raises(ValueError):
        cursor.set_stream(-1, StringIO())
    cursor.set_stream(0, StringIO())
    row = cursor.fetchone()
    assert isinstance(row[0], StringIO)
    assert row[0].getvalue() == val

    # test nvarchar(max) with NULL value
    cursor.execute("select cast(NULL as nvarchar(max)), 1".format(val))
    cursor.set_stream(0, StringIO())
    row = cursor.fetchone()
    assert row[0] is None

    # test varchar(max)
    cursor.execute("select '{}', 1".format(val))
    cursor.set_stream(0, StringIO())
    row = cursor.fetchone()
    assert isinstance(row[0], StringIO)
    assert row[0].getvalue() == val

    # test varbinary(max)
    cursor.execute("select cast('{}' as varbinary(max)), 1".format(val))
    cursor.set_stream(0, BytesIO())
    row = cursor.fetchone()
    assert isinstance(row[0], BytesIO)
    assert row[0].getvalue().decode('ascii') == val

    # test image type
    cursor.execute("select cast('{}' as image), 1".format(val))
    cursor.set_stream(0, BytesIO())
    row = cursor.fetchone()
    assert isinstance(row[0], BytesIO)
    assert row[0].getvalue().decode('ascii') == val

    # test ntext type
    cursor.execute("select cast('{}' as ntext), 1".format(val))
    cursor.set_stream(0, StringIO())
    row = cursor.fetchone()
    assert isinstance(row[0], StringIO)
    assert row[0].getvalue() == val

    # test text type
    cursor.execute("select cast('{}' as text), 1".format(val))
    cursor.set_stream(0, StringIO())
    row = cursor.fetchone()
    assert isinstance(row[0], StringIO)
    assert row[0].getvalue() == val

    # test xml type
    xml_val = '<root>{}</root>'.format(val)
    cursor.execute("select cast('{}' as xml), 1".format(xml_val))
    cursor.set_stream(0, StringIO())
    row = cursor.fetchone()
    assert isinstance(row[0], StringIO)
    assert row[0].getvalue() == xml_val


def test_dictionary_params(cursor):
    assert cursor.execute_scalar('select %(param)s', {'param': None}) == None
    assert cursor.execute_scalar('select %(param)s', {'param': 1}) == 1


def test_properties(separate_db_connection):
    conn = separate_db_connection
    # this property is provided for compatibility with pymssql
    assert conn.autocommit_state == conn.autocommit
    # test set_autocommit which is provided for compatibility with ADO dbapi
    conn.set_autocommit(conn.autocommit)
    # test isolation_level property read/write
    conn.isolation_level = conn.isolation_level
    # test product_version property read
    logger.info("Product version %s", conn.product_version)
    conn.as_dict = conn.as_dict


def test_fetch_on_empty_dataset(cursor):
    cursor.execute('declare @x int')
    with pytest.raises(pytds.ProgrammingError):
        cursor.fetchall()


def test_isolation_level(separate_db_connection):
    conn = separate_db_connection
    # enable autocommit and then reenable to force new transaction to be started
    conn.autocommit = True
    conn.isolation_level = pytds.extensions.ISOLATION_LEVEL_SERIALIZABLE
    conn.autocommit = False
    with conn.cursor() as cur:
        cur.execute('select transaction_isolation_level '
                    'from sys.dm_exec_sessions where session_id = @@SPID')
        lvl, = cur.fetchone()
    assert pytds.extensions.ISOLATION_LEVEL_SERIALIZABLE == lvl


def test_bad_collation(cursor):
    # exception can be different
    with pytest.raises(UnicodeDecodeError):
        cursor.execute_scalar('select cast(0x90 as varchar)')
    # check that connection is still usable
    assert 1 == cursor.execute_scalar('select 1')


def test_overlimit(cursor):
    def test_val(val):
        cursor.execute('select %s', (val,))
        assert cursor.fetchone() == (val,)
        assert cursor.fetchone() is None

    ##cur.execute('select %s', '\x00'*(2**31))
    with pytest.raises(pytds.DataError):
        test_val(Decimal('1' + '0' * 38))
    with pytest.raises(pytds.DataError):
        test_val(Decimal('-1' + '0' * 38))
    with pytest.raises(pytds.DataError):
        test_val(Decimal('1E38'))
    val = -10 ** 38
    cursor.execute('select %s', (val,))
    assert cursor.fetchone() == (str(val),)
    assert cursor.fetchone() is None


def test_description(cursor):
    cursor.execute('select cast(12.65 as decimal(4,2)) as testname')
    assert cursor.description[0][0] == 'testname'
    assert cursor.description[0][1] == pytds.DECIMAL
    assert cursor.description[0][4] == 4
    assert cursor.description[0][5] == 2


def test_bug4(separate_db_connection):
    with separate_db_connection.cursor() as cursor:
        cursor.execute('''
        set transaction isolation level read committed
        select 1
        ''')
        assert cursor.fetchall() == [(1,)]


def test_row_strategies(separate_db_connection):
    conn = separate_db_connection
    conn.as_dict = True
    with conn.cursor() as cur:
        cur.execute('select 1 as f')
        assert cur.fetchall() == [{'f': 1}]
    conn.as_dict = False
    with conn.cursor() as cur:
        cur.execute('select 1 as f')
        assert cur.fetchall() == [(1,)]


def test_fetchone(cursor):
    cur = cursor
    cur.execute('select 10; select 12')
    assert (10,) == cur.fetchone()
    assert cur.nextset()
    assert (12,) == cur.fetchone()
    assert not cur.nextset()


def test_fetchall(cursor):
    cur = cursor
    cur.execute('select 10; select 12')
    assert [(10,)] == cur.fetchall()
    assert cur.nextset()
    assert [(12,)] == cur.fetchall()
    assert not cur.nextset()


def test_cursor_closing(db_connection):
    with db_connection.cursor() as cur:
        cur.execute('select 10; select 12')
        cur.fetchone()
    with db_connection.cursor() as cur2:
        cur2.execute('select 20')
        cur2.fetchone()


def test_multi_packet(cursor):
    cur = cursor
    param = 'x' * (cursor._conn()._conn.main_session._writer.bufsize * 3)
    cur.execute('select %s', (param,))
    assert [(param, )] == cur.fetchall()


def test_big_request(cursor):
    cur = cursor
    param = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(5000))
    params = (10, datetime.datetime(2012, 11, 19, 1, 21, 37, 3000), param, 'test')
    cur.execute('select %s, %s, %s, %s', params)
    assert [params] == cur.fetchall()


def test_row_count(cursor):
    cur = cursor
    cur.execute('''
    create table testtable_row_cnt (field int)
    ''')
    cur.execute('insert into testtable_row_cnt (field) values (1)')
    assert cur.rowcount == 1
    cur.execute('insert into testtable_row_cnt (field) values (2)')
    assert cur.rowcount == 1
    cur.execute('select * from testtable_row_cnt')
    cur.fetchall()
    assert cur.rowcount == 2


def test_no_rows(cursor):
    cur = cursor
    cur.execute('''
    create table testtable_no_rows (field int)
    ''')
    cur.execute('select * from testtable_no_rows')
    assert [] == cur.fetchall()


def test_fixed_size_data(cursor):
    cur = cursor
    cur.execute('''
    create table testtable_fixed_size_dt (chr char(5), nchr nchar(5), bfld binary(5))
    insert into testtable_fixed_size_dt values ('1', '2', cast('3' as binary(5)))
    ''')
    cur.execute('select * from testtable_fixed_size_dt')
    assert cur.fetchall() == [('1    ', '2    ', b'3\x00\x00\x00\x00')]


def test_transactions(separate_db_connection):
    conn = separate_db_connection
    conn.autocommit = False
    with conn.cursor() as cur:
        cur.execute('''
        create table testtable_trans (field datetime)
        ''')
        cur.execute("select object_id('testtable_trans')")
        assert (None,) != cur.fetchone()
        assert 1 == conn._trancount()
        conn.rollback()
        assert 1 == conn._trancount()
        cur.execute("select object_id('testtable_trans')")
        assert (None,) == cur.fetchone()

        cur.execute('''
        create table testtable_trans (field datetime)
        ''')

        conn.commit()

        cur.execute("select object_id('testtable_trans')")
        assert (None,) != cur.fetchone()

    with conn.cursor() as cur:
        cur.execute('''
        if object_id('testtable_trans') is not null
            drop table testtable_trans
        ''')
    conn.commit()


def test_manual_commit(separate_db_connection):
    conn = separate_db_connection
    conn.autocommit = False
    cur = conn.cursor()
    cur.execute("create table tbl(x int)")
    assert conn._conn.tds72_transaction
    try:
        cur.execute("create table tbl(x int)")
    except:
        pass
    trancount = cur.execute_scalar("select @@trancount")
    assert 1 == trancount, 'Should be in transaction even after errors'

    cur.execute("create table tbl(x int)")
    try:
        cur.execute("create table tbl(x int)")
    except:
        pass
    cur.callproc('sp_executesql', ('select @@trancount',))
    trancount, = cur.fetchone()
    assert 1 == trancount, 'Should be in transaction even after errors'


def test_closing_cursor_in_context(db_connection):
    with db_connection.cursor() as cur:
        cur.close()


def test_cursor_connection_property(db_connection):
    with db_connection.cursor() as cur:
        assert cur.connection is db_connection


def test_outparam_and_result_set(cursor):
    """
    Test stored procedure which has output parameters and also result set
    """
    cur = cursor
    logger.info('creating stored procedure')
    cur.execute('''
    CREATE PROCEDURE P_OutParam_ResultSet(@A INT OUTPUT)
    AS BEGIN
    SET @A = 3;
    SELECT 4 AS C;
    SELECT 5 AS C;
    END;
    '''
                )
    logger.info('executing stored procedure')
    cur.callproc('P_OutParam_ResultSet', [pytds.output(value=1)])
    assert [(4,)] == cur.fetchall()
    assert [3] == cur.get_proc_outputs()
    logger.info('execurint query after stored procedure')
    cur.execute('select 5')
    assert [(5,)] == cur.fetchall()


def test_outparam_null_default(cursor):
    with pytest.raises(ValueError):
        pytds.output(None, None)

    cur = cursor
    cur.execute('''
    create procedure outparam_null_testproc (@inparam int, @outint int = 8 output, @outstr varchar(max) = 'defstr' output)
    as
    begin
        set nocount on
        set @outint = isnull(@outint, -10) + @inparam
        set @outstr = isnull(@outstr, 'null') + cast(@inparam as varchar(max))
        set @inparam = 8
    end
    ''')
    values = cur.callproc('outparam_null_testproc', (1, pytds.output(value=4), pytds.output(value='str')))
    assert [1, 5, 'str1'] == values
    values = cur.callproc('outparam_null_testproc', (1, pytds.output(value=None, param_type='int'), pytds.output(value=None, param_type='varchar(max)')))
    assert [1, -9, 'null1'] == values
    values = cur.callproc('outparam_null_testproc', (1, pytds.output(value=pytds.default, param_type='int'), pytds.output(value=pytds.default, param_type='varchar(max)')))
    assert [1, 9, 'defstr1'] == values
    values = cur.callproc('outparam_null_testproc', (1, pytds.output(value=pytds.default, param_type='bit'), pytds.output(value=pytds.default, param_type='varchar(5)')))
    assert [1, 1, 'defst'] == values
    values = cur.callproc('outparam_null_testproc', (1, pytds.output(value=pytds.default, param_type=int), pytds.output(value=pytds.default, param_type=str)))
    assert [1, 9, 'defstr1'] == values


def test_invalid_ntlm_creds():
    if not LIVE_TEST:
        pytest.skip('LIVE_TEST is not set')
    with pytest.raises(pytds.OperationalError):
        pytds.connect(settings.HOST, auth=pytds.login.NtlmAuth(user_name='bad', password='bad'))


def test_open_with_different_blocksize():
    if not LIVE_TEST:
        pytest.skip('LIVE_TEST is not set')
    kwargs = settings.CONNECT_KWARGS.copy()
    # test very small block size
    kwargs['blocksize'] = 100
    with pytds.connect(*settings.CONNECT_ARGS, **kwargs):
        pass
    # test very large block size
    kwargs['blocksize'] = 1000000
    with pytds.connect(*settings.CONNECT_ARGS, **kwargs):
        pass


def test_nvarchar_multiple_rows(cursor):
    cursor.execute('''
    set nocount on
    declare @tbl table (id int primary key, fld nvarchar(max))
    insert into @tbl values(1, 'foo')
    insert into @tbl values(2, 'bar')
    select fld from @tbl order by id
    '''
    )
    assert cursor.fetchall() == [('foo',), ('bar',)]


def test_no_metadata_request(cursor):
    cursor._session.submit_rpc(
        rpc_name=pytds.tds_base.SP_PREPARE,
        params=cursor._session._convert_params((pytds.output(param_type=int), '@p1 int', 'select @p1')),
    )
    cursor._session.process_rpc()
    while cursor.nextset():
        pass
    res = cursor.get_proc_outputs()
    handle = res[0]
    logger.info("got handle %s", handle)
    cursor._session.submit_rpc(
        rpc_name=pytds.tds_base.SP_EXECUTE,
        params=cursor._session._convert_params((handle, 1)),
    )
    cursor._session.process_rpc()
    cursor._setup_row_factory()
    assert cursor.fetchall() == [(1,)]
    while cursor.nextset():
        pass
    cursor._session.submit_rpc(
        rpc_name=pytds.tds_base.SP_EXECUTE,
        params=cursor._session._convert_params((handle, 2)),
        flags=0x02  # no metadata
    )
    cursor._session.process_rpc()
    cursor._setup_row_factory()
    # for some reason SQL server still sends metadata back
    assert cursor.fetchall() == [(2,)]
    while cursor.nextset():
        pass


def test_with_sso():
    if not LIVE_TEST:
        pytest.skip('LIVE_TEST is not set')
    with pytds.connect(settings.HOST, use_sso=True) as conn:
        with conn.cursor() as cursor:
            cursor.execute('select 1')
            cursor.fetchall()



# simple interactive shell for MSSQL server
import pytds
import os


def main():
    conn = pytds.connect(dsn=os.getenv("HOST", "localhost"),
                         user=os.getenv("SQLUSER", "sa"),
                         password=os.getenv("SQLPASSWORD"),
                         cafile="/Users/denisenk/opensource/pytds/ca.pem",
                         enc_login_only=True)
    while True:
        try:
            sql = input("sql> ")
        except KeyboardInterrupt:
            return
        with conn.cursor() as cursor:
            try:
                cursor.execute(sql)
            except pytds.ProgrammingError as e:
                print("Error: " + str(e))
            else:
                for _, msg in cursor.messages:
                    print(msg.text)
                if cursor.description:
                    print('\t'.join(col[0] for col in cursor.description))
                    print('-' * 80)
                    count = 0
                    for row in cursor:
                        print('\t'.join(str(col) for col in row))
                        count += 1
                    print('-' * 80)
                    print("Returned {} rows".format(count))
                print()


main()


#: Transaction can read uncommitted data
ISOLATION_LEVEL_READ_UNCOMMITTED = 1

#: Transaction can read only committed data, will block on attempt
#: to read modified uncommitted data
ISOLATION_LEVEL_READ_COMMITTED = 2

#: Transaction will place lock on read records, other transactions
#: will block trying to modify such records
ISOLATION_LEVEL_REPEATABLE_READ = 3

#: Transaction will lock tables to prevent other transactions
#: from inserting new data that would match selected recordsets
ISOLATION_LEVEL_SERIALIZABLE = 4

#: Allows non-blocking consistent reads on a snapshot for transaction without
#: blocking other transactions changes
ISOLATION_LEVEL_SNAPSHOT = 5



import struct
import cProfile
import pstats
import io

import pytds.smp

transport = None

bufsize = 512
smp_header = struct.Struct('<BBHLLL')


class Sock:
    def __init__(self):
        self._read_pos = 0
        self._seq = 1
        self._buf = bytearray(b'\x00' * bufsize)
        smp_header.pack_into(self._buf, 0, 0x53, 0x8, 0, bufsize, self._seq, 4)

    def sendall(self, data, flags=0):
        pass

    def recv_into(self, buffer, size=0):
        if size == 0:
            size = len(buffer)
        res = self.recv(size)
        buffer[:len(res)] = res[:]
        return len(res)

    def recv(self, size):
        if self._read_pos >= len(self._buf):
            self._seq += 1
            smp_header.pack_into(self._buf, 0, 0x53, 0x8, 0, bufsize, self._seq, 4)
            self._read_pos = 0
        res = self._buf[self._read_pos:self._read_pos + size]
        self._read_pos += len(res)
        return res

    def close(self):
        pass


sock = Sock()
mgr = pytds.smp.SmpManager(transport=sock)
sess = mgr.create_session()
pr = cProfile.Profile()
pr.enable()
buf = bytearray(b'\x00' * bufsize)
for _ in range(50000):
    sess.recv_into(buf)
pr.disable()
sortby = 'tottime'
ps = pstats.Stats(pr).sort_stats(sortby)
ps.print_stats()



import codecs
import struct

TDS_CHARSET_ISO_8859_1 = 1
TDS_CHARSET_CP1251 = 2
TDS_CHARSET_CP1252 = 3
TDS_CHARSET_UCS_2LE = 4

TDS_CHARSET_UNICODE = 5

ucs2_codec = codecs.lookup('utf_16_le')


def sortid2charset(sort_id):
    sql_collate = sort_id
    #
    # The table from the MSQLServer reference "Windows Collation Designators"
    # and from " NLS Information for Microsoft Windows XP"
    #
    if sql_collate in (
            30,  # SQL_Latin1_General_CP437_BIN
            31,  # SQL_Latin1_General_CP437_CS_AS
            32,  # SQL_Latin1_General_CP437_CI_AS
            33,  # SQL_Latin1_General_Pref_CP437_CI_AS
            34):  # SQL_Latin1_General_CP437_CI_AI
        return 'CP437'
    elif sql_collate in (
            40,  # SQL_Latin1_General_CP850_BIN
            41,  # SQL_Latin1_General_CP850_CS_AS
            42,  # SQL_Latin1_General_CP850_CI_AS
            43,  # SQL_Latin1_General_Pref_CP850_CI_AS
            44,  # SQL_Latin1_General_CP850_CI_AI
            49,  # SQL_1xCompat_CP850_CI_AS
            55,  # SQL_AltDiction_CP850_CS_AS
            56,  # SQL_AltDiction_Pref_CP850_CI_AS
            57,  # SQL_AltDiction_CP850_CI_AI
            58,  # SQL_Scandinavian_Pref_CP850_CI_AS
            59,  # SQL_Scandinavian_CP850_CS_AS
            60,  # SQL_Scandinavian_CP850_CI_AS
            61):  # SQL_AltDiction_CP850_CI_AS
        return 'CP850'
    elif sql_collate in (
            80,  # SQL_Latin1_General_1250_BIN
            81,  # SQL_Latin1_General_CP1250_CS_AS
            82,  # SQL_Latin1_General_Cp1250_CI_AS_KI_WI
            83,  # SQL_Czech_Cp1250_CS_AS_KI_WI
            84,  # SQL_Czech_Cp1250_CI_AS_KI_WI
            85,  # SQL_Hungarian_Cp1250_CS_AS_KI_WI
            86,  # SQL_Hungarian_Cp1250_CI_AS_KI_WI
            87,  # SQL_Polish_Cp1250_CS_AS_KI_WI
            88,  # SQL_Polish_Cp1250_CI_AS_KI_WI
            89,  # SQL_Romanian_Cp1250_CS_AS_KI_WI
            90,  # SQL_Romanian_Cp1250_CI_AS_KI_WI
            91,  # SQL_Croatian_Cp1250_CS_AS_KI_WI
            92,  # SQL_Croatian_Cp1250_CI_AS_KI_WI
            93,  # SQL_Slovak_Cp1250_CS_AS_KI_WI
            94,  # SQL_Slovak_Cp1250_CI_AS_KI_WI
            95,  # SQL_Slovenian_Cp1250_CS_AS_KI_WI
            96,  # SQL_Slovenian_Cp1250_CI_AS_KI_WI
            ):
        return 'CP1250'
    elif sql_collate in (
            104,  # SQL_Latin1_General_1251_BIN
            105,  # SQL_Latin1_General_CP1251_CS_AS
            106,  # SQL_Latin1_General_CP1251_CI_AS
            107,  # SQL_Ukrainian_Cp1251_CS_AS_KI_WI
            108,  # SQL_Ukrainian_Cp1251_CI_AS_KI_WI
            ):
        return 'CP1251'
    elif sql_collate in (
            51,  # SQL_Latin1_General_Cp1_CS_AS_KI_WI
            52,  # SQL_Latin1_General_Cp1_CI_AS_KI_WI
            53,  # SQL_Latin1_General_Pref_Cp1_CI_AS_KI_WI
            54,  # SQL_Latin1_General_Cp1_CI_AI_KI_WI
            183,  # SQL_Danish_Pref_Cp1_CI_AS_KI_WI
            184,  # SQL_SwedishPhone_Pref_Cp1_CI_AS_KI_WI
            185,  # SQL_SwedishStd_Pref_Cp1_CI_AS_KI_WI
            186,  # SQL_Icelandic_Pref_Cp1_CI_AS_KI_WI
            ):
        return 'CP1252'
    elif sql_collate in (
            112,  # SQL_Latin1_General_1253_BIN
            113,  # SQL_Latin1_General_CP1253_CS_AS
            114,  # SQL_Latin1_General_CP1253_CI_AS
            120,  # SQL_MixDiction_CP1253_CS_AS
            121,  # SQL_AltDiction_CP1253_CS_AS
            122,  # SQL_AltDiction2_CP1253_CS_AS
            124,  # SQL_Latin1_General_CP1253_CI_AI
            ):
        return 'CP1253'
    elif sql_collate in (
            128,  # SQL_Latin1_General_1254_BIN
            129,  # SQL_Latin1_General_Cp1254_CS_AS_KI_WI
            130,  # SQL_Latin1_General_Cp1254_CI_AS_KI_WI
            ):
        return 'CP1254'
    elif sql_collate in (
            136,  # SQL_Latin1_General_1255_BIN
            137,  # SQL_Latin1_General_CP1255_CS_AS
            138,  # SQL_Latin1_General_CP1255_CI_AS
            ):
        return 'CP1255'
    elif sql_collate in (
            144,  # SQL_Latin1_General_1256_BIN
            145,  # SQL_Latin1_General_CP1256_CS_AS
            146,  # SQL_Latin1_General_CP1256_CI_AS
            ):
        return 'CP1256'
    elif sql_collate in (
            152,  # SQL_Latin1_General_1257_BIN
            153,  # SQL_Latin1_General_CP1257_CS_AS
            154,  # SQL_Latin1_General_CP1257_CI_AS
            155,  # SQL_Estonian_Cp1257_CS_AS_KI_WI
            156,  # SQL_Estonian_Cp1257_CI_AS_KI_WI
            157,  # SQL_Latvian_Cp1257_CS_AS_KI_WI
            158,  # SQL_Latvian_Cp1257_CI_AS_KI_WI
            159,  # SQL_Lithuanian_Cp1257_CS_AS_KI_WI
            160,  # SQL_Lithuanian_Cp1257_CI_AS_KI_WI
            ):
        return 'CP1257'
    else:
        raise Exception("Invalid collation: 0x%X" % (sql_collate, ))


def lcid2charset(lcid):
    if lcid in (0x405,
                0x40e,  # 0x1040e
                0x415, 0x418, 0x41a, 0x41b, 0x41c, 0x424,
                # 0x81a, seem wrong in XP table TODO check
                0x104e):
                    return 'CP1250'
    elif lcid in (0x402, 0x419, 0x422, 0x423, 0x42f, 0x43f,
                  0x440, 0x444, 0x450,
                  0x81a,  # ??
                  0x82c, 0x843, 0xc1a):
                    return 'CP1251'
    elif lcid in (0x1007, 0x1009, 0x100a, 0x100c, 0x1407,
                  0x1409, 0x140a, 0x140c, 0x1809, 0x180a,
                  0x180c, 0x1c09, 0x1c0a, 0x2009, 0x200a,
                  0x2409, 0x240a, 0x2809, 0x280a, 0x2c09,
                  0x2c0a, 0x3009, 0x300a, 0x3409, 0x340a,
                  0x380a, 0x3c0a, 0x400a, 0x403, 0x406,
                  0x407,  # 0x10407
                  0x409, 0x40a, 0x40b, 0x40c, 0x40f, 0x410,
                  0x413, 0x414, 0x416, 0x41d, 0x421, 0x42d,
                  0x436,
                  0x437,  # 0x10437
                  0x438,
                  # 0x439,  ??? Unicode only
                  0x43e, 0x440a, 0x441, 0x456, 0x480a,
                  0x4c0a, 0x500a, 0x807, 0x809, 0x80a,
                  0x80c, 0x810, 0x813, 0x814, 0x816,
                  0x81d, 0x83e, 0xc07, 0xc09, 0xc0a, 0xc0c):
            return 'CP1252'
    elif lcid == 0x408:
        return 'CP1253'
    elif lcid in (0x41f, 0x42c, 0x443):
        return 'CP1254'
    elif lcid == 0x40d:
        return 'CP1255'
    elif lcid in (0x1001, 0x1401, 0x1801, 0x1c01, 0x2001,
                  0x2401, 0x2801, 0x2c01, 0x3001, 0x3401,
                  0x3801, 0x3c01, 0x4001, 0x401, 0x420,
                  0x429, 0x801, 0xc01):
        return 'CP1256'
    elif lcid in (0x425, 0x426, 0x427,
                  0x827):  # ??
        return 'CP1257'
    elif lcid == 0x42a:
        return 'CP1258'
    elif lcid == 0x41e:
        return 'CP874'
    elif lcid == 0x411:  # 0x10411
        return 'CP932'
    elif lcid in (0x1004,
                  0x804):  # 0x20804
        return 'CP936'
    elif lcid == 0x412:  # 0x10412
        return 'CP949'
    elif lcid in (0x1404,
                  0x404,  # 0x30404
                  0xc04):
        return 'CP950'
    else:
        return 'CP1252'


class Collation(object):
    _coll_struct = struct.Struct('<LB')
    wire_size = _coll_struct.size
    f_ignore_case = 0x100000
    f_ignore_accent = 0x200000
    f_ignore_width = 0x400000
    f_ignore_kana = 0x800000
    f_binary = 0x1000000
    f_binary2 = 0x2000000

    def __init__(self, lcid, sort_id, ignore_case, ignore_accent, ignore_width, ignore_kana, binary, binary2, version):
        self.lcid = lcid
        self.sort_id = sort_id
        self.ignore_case = ignore_case
        self.ignore_accent = ignore_accent
        self.ignore_width = ignore_width
        self.ignore_kana = ignore_kana
        self.binary = binary
        self.binary2 = binary2
        self.version = version

    def __repr__(self):
        fmt = 'Collation(lcid={0}, sort_id={1}, ignore_case={2}, ignore_accent={3}, ignore_width={4},' \
               ' ignore_kana={5}, binary={6}, binary2={7}, version={8})'
        return fmt.format(
            self.lcid,
            self.sort_id,
            self.ignore_case,
            self.ignore_accent,
            self.ignore_width,
            self.ignore_kana,
            self.binary,
            self.binary2,
            self.version
        )

    @classmethod
    def unpack(cls, b):
        lump, sort_id = cls._coll_struct.unpack_from(b)
        lcid = lump & 0xfffff
        ignore_case = bool(lump & cls.f_ignore_case)
        ignore_accent = bool(lump & cls.f_ignore_accent)
        ignore_width = bool(lump & cls.f_ignore_width)
        ignore_kana = bool(lump & cls.f_ignore_kana)
        binary = bool(lump & cls.f_binary)
        binary2 = bool(lump & cls.f_binary2)
        version = (lump & 0xf0000000) >> 26
        return cls(lcid=lcid,
                   ignore_case=ignore_case,
                   ignore_accent=ignore_accent,
                   ignore_width=ignore_width,
                   ignore_kana=ignore_kana,
                   binary=binary,
                   binary2=binary2,
                   version=version,
                   sort_id=sort_id)

    def pack(self):
        lump = 0
        lump |= self.lcid & 0xfffff
        lump |= (self.version << 26) & 0xf0000000
        if self.ignore_case:
            lump |= self.f_ignore_case
        if self.ignore_accent:
            lump |= self.f_ignore_accent
        if self.ignore_width:
            lump |= self.f_ignore_width
        if self.ignore_kana:
            lump |= self.f_ignore_kana
        if self.binary:
            lump |= self.f_binary
        if self.binary2:
            lump |= self.f_binary2
        return self._coll_struct.pack(lump, self.sort_id)

    def get_charset(self):
        if self.sort_id:
            return sortid2charset(self.sort_id)
        else:
            return lcid2charset(self.lcid)

    def get_codec(self):
        return codecs.lookup(self.get_charset())

    # TODO: define __repr__ and __unicode__


raw_collation = Collation(0, 0, 0, 0, 0, 0, 0, 0, 0)



import codecs
import contextlib
import logging
import datetime
import six
import socket
import struct

from .collate import ucs2_codec, Collation, lcid2charset, raw_collation
from . import tds_base
from . import tds_types
from . import tls
from .tds_base import readall, readall_fast, skipall, PreLoginEnc, PreLoginToken

logger = logging.getLogger()

# packet header
# https://msdn.microsoft.com/en-us/library/dd340948.aspx
_header = struct.Struct('>BBHHBx')

_byte = struct.Struct('B')
_smallint_le = struct.Struct('<h')
_smallint_be = struct.Struct('>h')
_usmallint_le = struct.Struct('<H')
_usmallint_be = struct.Struct('>H')
_int_le = struct.Struct('<l')
_int_be = struct.Struct('>l')
_uint_le = struct.Struct('<L')
_uint_be = struct.Struct('>L')
_int8_le = struct.Struct('<q')
_int8_be = struct.Struct('>q')
_uint8_le = struct.Struct('<Q')
_uint8_be = struct.Struct('>Q')

logging_enabled = False


# stored procedure output parameter
class output(object):
    @property
    def type(self):
        """
        This is either the sql type declaration or python type instance
        of the parameter.
        """
        return self._type

    @property
    def value(self):
        """
        This is the value of the parameter.
        """
        return self._value

    def __init__(self, value=None, param_type=None):
        """ Creates procedure output parameter.

        :param param_type: either sql type declaration or python type
        :param value: value to pass into procedure
        """
        if param_type is None:
            if value is None or value is default:
                raise ValueError('Output type cannot be autodetected')
        elif isinstance(param_type, type) and value is not None:
            if value is not default and not isinstance(value, param_type):
                raise ValueError('value should match param_type, value is {}, param_type is \'{}\''.format(repr(value), param_type.__name__))
        self._type = param_type
        self._value = value


class _Default(object):
    pass

default = _Default()


def tds7_crypt_pass(password):
    """ Mangle password according to tds rules

    :param password: Password str
    :returns: Byte-string with encoded password
    """
    encoded = bytearray(ucs2_codec.encode(password)[0])
    for i, ch in enumerate(encoded):
        encoded[i] = ((ch << 4) & 0xff | (ch >> 4)) ^ 0xA5
    return encoded


class _TdsEnv:
    def __init__(self):
        self.database = None
        self.language = None
        self.charset = None


class _TdsReader(object):
    """ TDS stream reader

    Provides stream-like interface for TDS packeted stream.
    Also provides convinience methods to decode primitive data like
    different kinds of integers etc.
    """
    def __init__(self, session):
        self._buf = bytearray(b'\x00' * 4096)
        self._bufview = memoryview(self._buf)
        self._pos = len(self._buf)  # position in the buffer
        self._have = 0  # number of bytes read from packet
        self._size = 0  # size of current packet
        self._session = session
        self._transport = session._transport
        self._type = None
        self._status = None

    def set_block_size(self, size):
        self._buf = bytearray(b'\x00' * size)
        self._bufview = memoryview(self._buf)

    def get_block_size(self):
        return len(self._buf)

    @property
    def session(self):
        """ Link to :class:`_TdsSession` object
        """
        return self._session

    @property
    def packet_type(self):
        """ Type of current packet

        Possible values are TDS_QUERY, TDS_LOGIN, etc.
        """
        return self._type

    def read_fast(self, size):
        """ Faster version of read

        Instead of returning sliced buffer it returns reference to internal
        buffer and the offset to this buffer.

        :param size: Number of bytes to read
        :returns: Tuple of bytes buffer, and offset in this buffer
        """
        if self._pos >= self._size:
            self._read_packet()
        offset = self._pos
        to_read = min(size, self._size - self._pos)
        self._pos += to_read
        return self._buf, offset

    def recv(self, size):
        if self._pos >= self._size:
            self._read_packet()
        offset = self._pos
        to_read = min(size, self._size - self._pos)
        self._pos += to_read
        return self._buf[offset:offset+to_read]

    def unpack(self, struc):
        """ Unpacks given structure from stream

        :param struc: A struct.Struct instance
        :returns: Result of unpacking
        """
        buf, offset = readall_fast(self, struc.size)
        return struc.unpack_from(buf, offset)

    def get_byte(self):
        """ Reads one byte from stream """
        return self.unpack(_byte)[0]

    def get_smallint(self):
        """ Reads 16bit signed integer from the stream """
        return self.unpack(_smallint_le)[0]

    def get_usmallint(self):
        """ Reads 16bit unsigned integer from the stream """
        return self.unpack(_usmallint_le)[0]

    def get_int(self):
        """ Reads 32bit signed integer from the stream """
        return self.unpack(_int_le)[0]

    def get_uint(self):
        """ Reads 32bit unsigned integer from the stream """
        return self.unpack(_uint_le)[0]

    def get_uint_be(self):
        """ Reads 32bit unsigned big-endian integer from the stream """
        return self.unpack(_uint_be)[0]

    def get_uint8(self):
        """ Reads 64bit unsigned integer from the stream """
        return self.unpack(_uint8_le)[0]

    def get_int8(self):
        """ Reads 64bit signed integer from the stream """
        return self.unpack(_int8_le)[0]

    def read_ucs2(self, num_chars):
        """ Reads num_chars UCS2 string from the stream """
        buf = readall(self, num_chars * 2)
        return ucs2_codec.decode(buf)[0]

    def read_str(self, size, codec):
        """ Reads byte string from the stream and decodes it

        :param size: Size of string in bytes
        :param codec: Instance of codec to decode string
        :returns: Unicode string
        """
        return codec.decode(readall(self, size))[0]

    def get_collation(self):
        """ Reads :class:`Collation` object from stream """
        buf = readall(self, Collation.wire_size)
        return Collation.unpack(buf)

    def _read_packet(self):
        """ Reads next TDS packet from the underlying transport

        If timeout is happened during reading of packet's header will
        cancel current request.
        Can only be called when transport's read pointer is at the begining
        of the packet.
        """
        try:
            pos = 0
            while pos < _header.size:
                received = self._transport.recv_into(self._bufview[pos:], _header.size - pos)
                if received == 0:
                    raise tds_base.ClosedConnectionError()
                pos += received
        except tds_base.TimeoutError:
            self._session.put_cancel()
            raise
        self._pos = _header.size
        self._type, self._status, self._size, self._session._spid, _ = _header.unpack_from(self._bufview, 0)
        self._have = pos
        while pos < self._size:
            received = self._transport.recv_into(self._bufview[pos:], self._size - pos)
            if received == 0:
                raise tds_base.ClosedConnectionError()
            pos += received
            self._have += received

    def read_whole_packet(self):
        """ Reads single packet and returns bytes payload of the packet

        Can only be called when transport's read pointer is at the beginning
        of the packet.
        """
        self._read_packet()
        return readall(self, self._size - _header.size)


class _TdsWriter(object):
    """ TDS stream writer

    Handles splitting of incoming data into TDS packets according to TDS protocol.
    Provides convinience methods for writing primitive data types.
    """
    def __init__(self, session, bufsize):
        self._session = session
        self._tds = session
        self._transport = session._transport
        self._pos = 0
        self._buf = bytearray(bufsize)
        self._packet_no = 0
        self._type = 0

    @property
    def session(self):
        """ Back reference to parent :class:`_TdsSession` object """
        return self._session

    @property
    def bufsize(self):
        """ Size of the buffer """
        return len(self._buf)

    @bufsize.setter
    def bufsize(self, bufsize):
        if len(self._buf) == bufsize:
            return

        if bufsize > len(self._buf):
            self._buf.extend(b'\0' * (bufsize - len(self._buf)))
        else:
            self._buf = self._buf[0:bufsize]

    def begin_packet(self, packet_type):
        """ Starts new packet stream

        :param packet_type: Type of TDS stream, e.g. TDS_PRELOGIN, TDS_QUERY etc.
        """
        self._type = packet_type
        self._pos = 8

    def pack(self, struc, *args):
        """ Packs and writes structure into stream """
        self.write(struc.pack(*args))

    def put_byte(self, value):
        """ Writes single byte into stream """
        self.pack(_byte, value)

    def put_smallint(self, value):
        """ Writes 16-bit signed integer into the stream """
        self.pack(_smallint_le, value)

    def put_usmallint(self, value):
        """ Writes 16-bit unsigned integer into the stream """
        self.pack(_usmallint_le, value)

    def put_usmallint_be(self, value):
        """ Writes 16-bit unsigned big-endian integer into the stream """
        self.pack(_usmallint_be, value)

    def put_int(self, value):
        """ Writes 32-bit signed integer into the stream """
        self.pack(_int_le, value)

    def put_uint(self, value):
        """ Writes 32-bit unsigned integer into the stream """
        self.pack(_uint_le, value)

    def put_uint_be(self, value):
        """ Writes 32-bit unsigned big-endian integer into the stream """
        self.pack(_uint_be, value)

    def put_int8(self, value):
        """ Writes 64-bit signed integer into the stream """
        self.pack(_int8_le, value)

    def put_uint8(self, value):
        """ Writes 64-bit unsigned integer into the stream """
        self.pack(_uint8_le, value)

    def put_collation(self, collation):
        """ Writes :class:`Collation` structure into the stream """
        self.write(collation.pack())

    def write(self, data):
        """ Writes given bytes buffer into the stream

        Function returns only when entire buffer is written
        """
        data_off = 0
        while data_off < len(data):
            left = len(self._buf) - self._pos
            if left <= 0:
                self._write_packet(final=False)
            else:
                to_write = min(left, len(data) - data_off)
                self._buf[self._pos:self._pos + to_write] = data[data_off:data_off + to_write]
                self._pos += to_write
                data_off += to_write

    def write_b_varchar(self, s):
        self.put_byte(len(s))
        self.write_ucs2(s)

    def write_ucs2(self, s):
        """ Write string encoding it in UCS2 into stream """
        self.write_string(s, ucs2_codec)

    def write_string(self, s, codec):
        """ Write string encoding it with codec into stream """
        for i in range(0, len(s), self.bufsize):
            chunk = s[i:i + self.bufsize]
            buf, consumed = codec.encode(chunk)
            assert consumed == len(chunk)
            self.write(buf)

    def flush(self):
        """ Closes current packet stream """
        return self._write_packet(final=True)

    def _write_packet(self, final):
        """ Writes single TDS packet into underlying transport.

        Data for the packet is taken from internal buffer.

        :param final: True means this is the final packet in substream.
        """
        status = 1 if final else 0
        _header.pack_into(self._buf, 0, self._type, status, self._pos, 0, self._packet_no)
        self._packet_no = (self._packet_no + 1) % 256
        self._transport.sendall(self._buf[:self._pos])
        self._pos = 8


def _create_exception_by_message(msg, custom_error_msg=None):
    msg_no = msg['msgno']
    if custom_error_msg is not None:
        error_msg = custom_error_msg
    else:
        error_msg = msg['message']
    if msg_no in tds_base.prog_errors:
        ex = tds_base.ProgrammingError(error_msg)
    elif msg_no in tds_base.integrity_errors:
        ex = tds_base.IntegrityError(error_msg)
    else:
        ex = tds_base.OperationalError(error_msg)
    ex.msg_no = msg['msgno']
    ex.text = msg['message']
    ex.srvname = msg['server']
    ex.procname = msg['proc_name']
    ex.number = msg['msgno']
    ex.severity = msg['severity']
    ex.state = msg['state']
    ex.line = msg['line_number']
    return ex


class _TdsSession(object):
    """ TDS session

    Represents a single TDS session within MARS connection, when MARS enabled there could be multiple TDS sessions
    within one connection.
    """
    def __init__(self, tds, transport, tzinfo_factory):
        self.out_pos = 8
        self.res_info = None
        self.in_cancel = False
        self.wire_mtx = None
        self.param_info = None
        self.has_status = False
        self.ret_status = None
        self.skipped_to_status = False
        self._transport = transport
        self._reader = _TdsReader(self)
        self._reader._transport = transport
        self._writer = _TdsWriter(self, tds.bufsize)
        self._writer._transport = transport
        self.in_buf_max = 0
        self.state = tds_base.TDS_IDLE
        self._tds = tds
        self.messages = []
        self.rows_affected = -1
        self.use_tz = tds.use_tz
        self._spid = 0
        self.tzinfo_factory = tzinfo_factory
        self.more_rows = False
        self.done_flags = 0
        self.internal_sp_called = 0
        self.output_params = {}
        self.authentication = None
        self.return_value_index = 0
        self._out_params_indexes = []
        self.row = None
        self.end_marker = 0

    def log_response_message(self, msg):
        # logging is disabled by default
        if logging_enabled:
            logger.info('[%d] %s', self._spid, msg)

    def __repr__(self):
        fmt = "<_TdsSession state={} tds={} messages={} rows_affected={} use_tz={} spid={} in_cancel={}>"
        res = fmt.format(repr(self.state), repr(self._tds), repr(self.messages),
                         repr(self.rows_affected), repr(self.use_tz), repr(self._spid),
                         self.in_cancel)
        return res

    def raise_db_exception(self):
        """ Raises exception from last server message

        This function will skip messages: The statement has been terminated
        """
        if not self.messages:
            raise tds_base.Error("Request failed, server didn't send error message")
        msg = None
        while True:
            msg = self.messages[-1]
            if msg['msgno'] == 3621:  # the statement has been terminated
                self.messages = self.messages[:-1]
            else:
                break

        error_msg = ' '.join(m['message'] for m in self.messages)
        ex = _create_exception_by_message(msg, error_msg)
        raise ex

    def get_type_info(self, curcol):
        """ Reads TYPE_INFO structure (http://msdn.microsoft.com/en-us/library/dd358284.aspx)

        :param curcol: An instance of :class:`Column` that will receive read information
        """
        r = self._reader
        # User defined data type of the column
        if tds_base.IS_TDS72_PLUS(self):
            user_type = r.get_uint()
        else:
            user_type = r.get_usmallint()
        curcol.column_usertype = user_type
        curcol.flags = r.get_usmallint()  # Flags
        type_id = r.get_byte()
        serializer_class = self._tds.type_factory.get_type_serializer(type_id)
        curcol.serializer = serializer_class.from_stream(r)

    def tds7_process_result(self):
        """ Reads and processes COLMETADATA stream

        This stream contains a list of returned columns.
        Stream format link: http://msdn.microsoft.com/en-us/library/dd357363.aspx
        """
        self.log_response_message('got COLMETADATA')
        r = self._reader

        # read number of columns and allocate the columns structure

        num_cols = r.get_smallint()

        # This can be a DUMMY results token from a cursor fetch

        if num_cols == -1:
            return

        self.param_info = None
        self.has_status = False
        self.ret_status = None
        self.skipped_to_status = False
        self.rows_affected = tds_base.TDS_NO_COUNT
        self.more_rows = True
        self.row = [None] * num_cols
        self.res_info = info = _Results()

        #
        # loop through the columns populating COLINFO struct from
        # server response
        #
        header_tuple = []
        for col in range(num_cols):
            curcol = tds_base.Column()
            info.columns.append(curcol)
            self.get_type_info(curcol)

            curcol.column_name = r.read_ucs2(r.get_byte())
            precision = curcol.serializer.precision
            scale = curcol.serializer.scale
            size = curcol.serializer.size
            header_tuple.append(
                (curcol.column_name,
                 curcol.serializer.get_typeid(),
                 None,
                 size,
                 precision,
                 scale,
                 curcol.flags & tds_base.Column.fNullable))
        info.description = tuple(header_tuple)
        return info

    def process_param(self):
        """ Reads and processes RETURNVALUE stream.

        This stream is used to send OUTPUT parameters from RPC to client.
        Stream format url: http://msdn.microsoft.com/en-us/library/dd303881.aspx
        """
        self.log_response_message('got RETURNVALUE message')
        r = self._reader
        if tds_base.IS_TDS72_PLUS(self):
            ordinal = r.get_usmallint()
        else:
            r.get_usmallint()  # ignore size
            ordinal = self._out_params_indexes[self.return_value_index]
        name = r.read_ucs2(r.get_byte())
        r.get_byte()  # 1 - OUTPUT of sp, 2 - result of udf
        param = tds_base.Column()
        param.column_name = name
        self.get_type_info(param)
        param.value = param.serializer.read(r)
        self.output_params[ordinal] = param
        self.return_value_index += 1

    def process_cancel(self):
        """
        Process the incoming token stream until it finds
        an end token DONE with the cancel flag set.
        At that point the connection should be ready to handle a new query.

        In case when no cancel request is pending this function does nothing.
        """
        self.log_response_message('got CANCEL message')
        # silly cases, nothing to do
        if not self.in_cancel:
            return

        while True:
            token_id = self.get_token_id()
            self.process_token(token_id)
            if not self.in_cancel:
                return

    def process_msg(self, marker):
        """ Reads and processes ERROR/INFO streams

        Stream formats:

        - ERROR: http://msdn.microsoft.com/en-us/library/dd304156.aspx
        - INFO: http://msdn.microsoft.com/en-us/library/dd303398.aspx

        :param marker: TDS_ERROR_TOKEN or TDS_INFO_TOKEN
        """
        self.log_response_message('got ERROR/INFO message')
        r = self._reader
        r.get_smallint()  # size
        msg = {'marker': marker, 'msgno': r.get_int(), 'state': r.get_byte(), 'severity': r.get_byte(),
               'sql_state': None}
        if marker == tds_base.TDS_INFO_TOKEN:
            msg['priv_msg_type'] = 0
        elif marker == tds_base.TDS_ERROR_TOKEN:
            msg['priv_msg_type'] = 1
        else:
            logger.error('tds_process_msg() called with unknown marker "{0}"'.format(marker))
        msg['message'] = r.read_ucs2(r.get_smallint())
        # server name
        msg['server'] = r.read_ucs2(r.get_byte())
        # stored proc name if available
        msg['proc_name'] = r.read_ucs2(r.get_byte())
        msg['line_number'] = r.get_int() if tds_base.IS_TDS72_PLUS(self) else r.get_smallint()
        # in case extended error data is sent, we just try to discard it

        # special case
        self.messages.append(msg)

    def process_row(self):
        """ Reads and handles ROW stream.

        This stream contains list of values of one returned row.
        Stream format url: http://msdn.microsoft.com/en-us/library/dd357254.aspx
        """
        self.log_response_message("got ROW message")
        r = self._reader
        info = self.res_info
        info.row_count += 1
        for i, curcol in enumerate(info.columns):
            curcol.value = self.row[i] = curcol.serializer.read(r)

    def process_nbcrow(self):
        """ Reads and handles NBCROW stream.

        This stream contains list of values of one returned row in a compressed way,
        introduced in TDS 7.3.B
        Stream format url: http://msdn.microsoft.com/en-us/library/dd304783.aspx
        """
        self.log_response_message("got NBCROW message")
        r = self._reader
        info = self.res_info
        if not info:
            self.bad_stream('got row without info')
        assert len(info.columns) > 0
        info.row_count += 1

        # reading bitarray for nulls, 1 represent null values for
        # corresponding fields
        nbc = readall(r, (len(info.columns) + 7) // 8)
        for i, curcol in enumerate(info.columns):
            if tds_base.my_ord(nbc[i // 8]) & (1 << (i % 8)):
                value = None
            else:
                value = curcol.serializer.read(r)
            self.row[i] = value

    def process_orderby(self):
        """ Reads and processes ORDER stream

        Used to inform client by which column dataset is ordered.
        Stream format url: http://msdn.microsoft.com/en-us/library/dd303317.aspx
        """
        r = self._reader
        skipall(r, r.get_smallint())

    def process_end(self, marker):
        """ Reads and processes DONE/DONEINPROC/DONEPROC streams

        Stream format urls:

        - DONE: http://msdn.microsoft.com/en-us/library/dd340421.aspx
        - DONEINPROC: http://msdn.microsoft.com/en-us/library/dd340553.aspx
        - DONEPROC: http://msdn.microsoft.com/en-us/library/dd340753.aspx

        :param marker: Can be TDS_DONE_TOKEN or TDS_DONEINPROC_TOKEN or TDS_DONEPROC_TOKEN
        """
        code_to_str = {
            tds_base.TDS_DONE_TOKEN: 'DONE',
            tds_base.TDS_DONEINPROC_TOKEN: 'DONEINPROC',
            tds_base.TDS_DONEPROC_TOKEN: 'DONEPROC',
        }
        self.end_marker = marker
        self.more_rows = False
        r = self._reader
        status = r.get_usmallint()
        r.get_usmallint()  # cur_cmd
        more_results = status & tds_base.TDS_DONE_MORE_RESULTS != 0
        was_cancelled = status & tds_base.TDS_DONE_CANCELLED != 0
        done_count_valid = status & tds_base.TDS_DONE_COUNT != 0
        if self.res_info:
            self.res_info.more_results = more_results
        rows_affected = r.get_int8() if tds_base.IS_TDS72_PLUS(self) else r.get_int()
        self.log_response_message("got {} message, more_res={}, cancelled={}, rows_affected={}".format(
            code_to_str[marker], more_results, was_cancelled, rows_affected))
        if was_cancelled or (not more_results and not self.in_cancel):
            self.in_cancel = False
            self.set_state(tds_base.TDS_IDLE)
        if done_count_valid:
            self.rows_affected = rows_affected
        else:
            self.rows_affected = -1
        self.done_flags = status
        if self.done_flags & tds_base.TDS_DONE_ERROR and not was_cancelled and not self.in_cancel:
            self.raise_db_exception()

    def process_env_chg(self):
        """ Reads and processes ENVCHANGE stream.

        Stream info url: http://msdn.microsoft.com/en-us/library/dd303449.aspx
        """
        self.log_response_message("got ENVCHANGE message")
        r = self._reader
        size = r.get_smallint()
        type_id = r.get_byte()
        if type_id == tds_base.TDS_ENV_SQLCOLLATION:
            size = r.get_byte()
            self.conn.collation = r.get_collation()
            logger.info('switched collation to %s', self.conn.collation)
            skipall(r, size - 5)
            # discard old one
            skipall(r, r.get_byte())
        elif type_id == tds_base.TDS_ENV_BEGINTRANS:
            size = r.get_byte()
            assert size == 8
            self.conn.tds72_transaction = r.get_uint8()
            skipall(r, r.get_byte())
        elif type_id == tds_base.TDS_ENV_COMMITTRANS or type_id == tds_base.TDS_ENV_ROLLBACKTRANS:
            self.conn.tds72_transaction = 0
            skipall(r, r.get_byte())
            skipall(r, r.get_byte())
        elif type_id == tds_base.TDS_ENV_PACKSIZE:
            newval = r.read_ucs2(r.get_byte())
            r.read_ucs2(r.get_byte())
            new_block_size = int(newval)
            if new_block_size >= 512:
                # Is possible to have a shrink if server limits packet
                # size more than what we specified
                #
                # Reallocate buffer if possible (strange values from server or out of memory) use older buffer */
                self._writer.bufsize = new_block_size
        elif type_id == tds_base.TDS_ENV_DATABASE:
            newval = r.read_ucs2(r.get_byte())
            logger.info('switched to database %s', newval)
            r.read_ucs2(r.get_byte())
            self.conn.env.database = newval
        elif type_id == tds_base.TDS_ENV_LANG:
            newval = r.read_ucs2(r.get_byte())
            logger.info('switched language to %s', newval)
            r.read_ucs2(r.get_byte())
            self.conn.env.language = newval
        elif type_id == tds_base.TDS_ENV_CHARSET:
            newval = r.read_ucs2(r.get_byte())
            logger.info('switched charset to %s', newval)
            r.read_ucs2(r.get_byte())
            self.conn.env.charset = newval
            remap = {'iso_1': 'iso8859-1'}
            self.conn.server_codec = codecs.lookup(remap.get(newval, newval))
        elif type_id == tds_base.TDS_ENV_DB_MIRRORING_PARTNER:
            newval = r.read_ucs2(r.get_byte())
            logger.info('got mirroring partner %s', newval)
            r.read_ucs2(r.get_byte())
        elif type_id == tds_base.TDS_ENV_LCID:
            lcid = int(r.read_ucs2(r.get_byte()))
            logger.info('switched lcid to %s', lcid)
            self.conn.server_codec = codecs.lookup(lcid2charset(lcid))
            r.read_ucs2(r.get_byte())
        elif type_id == tds_base.TDS_ENV_UNICODE_DATA_SORT_COMP_FLAGS:
            old_comp_flags = r.read_ucs2(r.get_byte())
            comp_flags = r.read_ucs2(r.get_byte())
            self.conn.comp_flags = comp_flags
        elif type_id == 20:
            # routing
            sz = r.get_usmallint()
            protocol = r.get_byte()
            protocol_property = r.get_usmallint()
            alt_server = r.read_ucs2(r.get_usmallint())
            logger.info('got routing info proto=%d proto_prop=%d alt_srv=%s', protocol, protocol_property, alt_server)
            self.conn.route = {
                'server': alt_server,
                'port': protocol_property,
            }
            # OLDVALUE = 0x00, 0x00
            r.get_usmallint()
        else:
            logger.warning("unknown env type: {0}, skipping".format(type_id))
            # discard byte values, not still supported
            skipall(r, size - 1)

    def process_auth(self):
        """ Reads and processes SSPI stream.

        Stream info: http://msdn.microsoft.com/en-us/library/dd302844.aspx
        """
        r = self._reader
        w = self._writer
        pdu_size = r.get_smallint()
        if not self.authentication:
            raise tds_base.Error('Got unexpected token')
        packet = self.authentication.handle_next(readall(r, pdu_size))
        if packet:
            w.write(packet)
            w.flush()

    def is_connected(self):
        """
        :return: True if transport is connected
        """
        return self._transport.is_connected()

    def bad_stream(self, msg):
        """ Called when input stream contains unexpected data.

        Will close stream and raise :class:`InterfaceError`
        :param msg: Message for InterfaceError exception.
        :return: Never returns, always raises exception.
        """
        self.close()
        raise tds_base.InterfaceError(msg)

    @property
    def tds_version(self):
        """ Returns integer encoded current TDS protocol version
        """
        return self._tds.tds_version

    @property
    def conn(self):
        """ Reference to owning :class:`_TdsSocket`
        """
        return self._tds

    def close(self):
        self._transport.close()

    def set_state(self, state):
        """ Switches state of the TDS session.

        It also does state transitions checks.
        :param state: New state, one of TDS_PENDING/TDS_READING/TDS_IDLE/TDS_DEAD/TDS_QUERING
        """
        prior_state = self.state
        if state == prior_state:
            return state
        if state == tds_base.TDS_PENDING:
            if prior_state in (tds_base.TDS_READING, tds_base.TDS_QUERYING):
                self.state = tds_base.TDS_PENDING
            else:
                raise tds_base.InterfaceError('logic error: cannot chage query state from {0} to {1}'.
                                              format(tds_base.state_names[prior_state], tds_base.state_names[state]))
        elif state == tds_base.TDS_READING:
            # transition to READING are valid only from PENDING
            if self.state != tds_base.TDS_PENDING:
                raise tds_base.InterfaceError('logic error: cannot change query state from {0} to {1}'.
                                              format(tds_base.state_names[prior_state], tds_base.state_names[state]))
            else:
                self.state = state
        elif state == tds_base.TDS_IDLE:
            if prior_state == tds_base.TDS_DEAD:
                raise tds_base.InterfaceError('logic error: cannot change query state from {0} to {1}'.
                                              format(tds_base.state_names[prior_state], tds_base.state_names[state]))
            self.state = state
        elif state == tds_base.TDS_DEAD:
            self.state = state
        elif state == tds_base.TDS_QUERYING:
            if self.state == tds_base.TDS_DEAD:
                raise tds_base.InterfaceError('logic error: cannot change query state from {0} to {1}'.
                                              format(tds_base.state_names[prior_state], tds_base.state_names[state]))
            elif self.state != tds_base.TDS_IDLE:
                raise tds_base.InterfaceError('logic error: cannot change query state from {0} to {1}'.
                                              format(tds_base.state_names[prior_state], tds_base.state_names[state]))
            else:
                self.rows_affected = tds_base.TDS_NO_COUNT
                self.internal_sp_called = 0
                self.state = state
        else:
            assert False
        return self.state

    @contextlib.contextmanager
    def querying_context(self, packet_type):
        """ Context manager for querying.

        Sets state to TDS_QUERYING, and reverts it to TDS_IDLE if exception happens inside managed block,
        and to TDS_PENDING if managed block succeeds and flushes buffer.
        """
        if self.set_state(tds_base.TDS_QUERYING) != tds_base.TDS_QUERYING:
            raise tds_base.Error("Couldn't switch to state")
        self._writer.begin_packet(packet_type)
        try:
            yield
        except:
            if self.state != tds_base.TDS_DEAD:
                self.set_state(tds_base.TDS_IDLE)
            raise
        else:
            self.set_state(tds_base.TDS_PENDING)
            self._writer.flush()

    def make_param(self, name, value):
        """ Generates instance of :class:`Column` from value and name

        Value can also be of a special types:

        - An instance of :class:`Column`, in which case it is just returned.
        - An instance of :class:`output`, in which case parameter will become
          an output parameter.
        - A singleton :var:`default`, in which case default value will be passed
          into a stored proc.

        :param name: Name of the parameter, will populate column_name property of returned column.
        :param value: Value of the parameter, also used to guess the type of parameter.
        :return: An instance of :class:`Column`
        """
        if isinstance(value, tds_base.Column):
            value.column_name = name
            return value
        column = tds_base.Column()
        column.column_name = name
        column.flags = 0
        
        if isinstance(value, output):
            column.flags |= tds_base.fByRefValue
            if isinstance(value.type, six.string_types):
                column.type = tds_types.sql_type_by_declaration(value.type)
            elif value.type:
                column.type = self.conn.type_inferrer.from_class(value.type)
            value = value.value

        if value is default:
            column.flags |= tds_base.fDefaultValue
            value = None

        column.value = value
        if column.type is None:
            column.type = self.conn.type_inferrer.from_value(value)
        return column

    def _convert_params(self, parameters):
        """ Converts a dict of list of parameters into a list of :class:`Column` instances.

        :param parameters: Can be a list of parameter values, or a dict of parameter names to values.
        :return: A list of :class:`Column` instances.
        """
        if isinstance(parameters, dict):
            return [self.make_param(name, value)
                    for name, value in parameters.items()]
        else:
            params = []
            for parameter in parameters:
                params.append(self.make_param('', parameter))
            return params

    def cancel_if_pending(self):
        """ Cancels current pending request.

        Does nothing if no request is pending, otherwise sends cancel request,
        and waits for response.
        """
        if self.state == tds_base.TDS_IDLE:
            return
        if not self.in_cancel:
            self.put_cancel()
        self.process_cancel()

    def submit_rpc(self, rpc_name, params, flags=0):
        """ Sends an RPC request.

        This call will transition session into pending state.
        If some operation is currently pending on the session, it will be
        cancelled before sending this request.

        Spec: http://msdn.microsoft.com/en-us/library/dd357576.aspx

        :param rpc_name: Name of the RPC to call, can be an instance of :class:`InternalProc`
        :param params: Stored proc parameters, should be a list of :class:`Column` instances.
        :param flags: See spec for possible flags.
        """
        logger.info('Sending RPC %s flags=%d', rpc_name, flags)
        self.messages = []
        self.output_params = {}
        self.cancel_if_pending()
        self.res_info = None
        w = self._writer
        with self.querying_context(tds_base.PacketType.RPC):
            if tds_base.IS_TDS72_PLUS(self):
                self._start_query()
            if tds_base.IS_TDS71_PLUS(self) and isinstance(rpc_name, tds_base.InternalProc):
                w.put_smallint(-1)
                w.put_smallint(rpc_name.proc_id)
            else:
                if isinstance(rpc_name, tds_base.InternalProc):
                    rpc_name = rpc_name.name
                w.put_smallint(len(rpc_name))
                w.write_ucs2(rpc_name)
            #
            # TODO support flags
            # bit 0 (1 as flag) in TDS7/TDS5 is "recompile"
            # bit 1 (2 as flag) in TDS7+ is "no metadata" bit this will prevent sending of column infos
            #
            w.put_usmallint(flags)
            self._out_params_indexes = []
            for i, param in enumerate(params):
                if param.flags & tds_base.fByRefValue:
                    self._out_params_indexes.append(i)
                w.put_byte(len(param.column_name))
                w.write_ucs2(param.column_name)
                #
                # TODO support other flags (use defaul null/no metadata)
                # bit 1 (2 as flag) in TDS7+ is "default value" bit
                # (what's the meaning of "default value" ?)
                #
                w.put_byte(param.flags)

                # TYPE_INFO structure: https://msdn.microsoft.com/en-us/library/dd358284.aspx
                serializer = param.choose_serializer(
                    type_factory=self._tds.type_factory,
                    collation=self._tds.collation or raw_collation
                )
                type_id = serializer.type
                w.put_byte(type_id)
                serializer.write_info(w)

                serializer.write(w, param.value)

    def submit_plain_query(self, operation):
        """ Sends a plain query to server.

        This call will transition session into pending state.
        If some operation is currently pending on the session, it will be
        cancelled before sending this request.

        Spec: http://msdn.microsoft.com/en-us/library/dd358575.aspx

        :param operation: A string representing sql statement.
        """
        self.messages = []
        self.cancel_if_pending()
        self.res_info = None
        logger.info("Sending query %s", operation[:100])
        w = self._writer
        with self.querying_context(tds_base.PacketType.QUERY):
            if tds_base.IS_TDS72_PLUS(self):
                self._start_query()
            w.write_ucs2(operation)

    def submit_bulk(self, metadata, rows):
        """ Sends insert bulk command.

        Spec: http://msdn.microsoft.com/en-us/library/dd358082.aspx

        :param metadata: A list of :class:`Column` instances.
        :param rows: A collection of rows, each row is a collection of values.
        :return:
        """
        logger.info('Sending INSERT BULK')
        num_cols = len(metadata)
        w = self._writer
        serializers = []
        with self.querying_context(tds_base.PacketType.BULK):
            w.put_byte(tds_base.TDS7_RESULT_TOKEN)
            w.put_usmallint(num_cols)
            for col in metadata:
                if tds_base.IS_TDS72_PLUS(self):
                    w.put_uint(col.column_usertype)
                else:
                    w.put_usmallint(col.column_usertype)
                w.put_usmallint(col.flags)
                serializer = col.choose_serializer(
                    type_factory=self._tds.type_factory,
                    collation=self._tds.collation,
                )
                type_id = serializer.type
                w.put_byte(type_id)
                serializers.append(serializer)
                serializer.write_info(w)
                w.put_byte(len(col.column_name))
                w.write_ucs2(col.column_name)
            for row in rows:
                w.put_byte(tds_base.TDS_ROW_TOKEN)
                for i, col in enumerate(metadata):
                    serializers[i].write(w, row[i])

            # https://msdn.microsoft.com/en-us/library/dd340421.aspx
            w.put_byte(tds_base.TDS_DONE_TOKEN)
            w.put_usmallint(tds_base.TDS_DONE_FINAL)
            w.put_usmallint(0)  # curcmd
            # row count
            if tds_base.IS_TDS72_PLUS(self):
                w.put_int8(0)
            else:
                w.put_int(0)

    def put_cancel(self):
        """ Sends a cancel request to the server.

        Switches connection to IN_CANCEL state.
        """
        logger.info('Sending CANCEL')
        self._writer.begin_packet(tds_base.PacketType.CANCEL)
        self._writer.flush()
        self.in_cancel = 1

    _begin_tran_struct_72 = struct.Struct('<HBB')

    def begin_tran(self, isolation_level=0):
        logger.info('Sending BEGIN TRAN il=%x', isolation_level)
        self.submit_begin_tran(isolation_level=isolation_level)
        self.process_simple_request()

    def submit_begin_tran(self, isolation_level=0):
        if tds_base.IS_TDS72_PLUS(self):
            self.messages = []
            self.cancel_if_pending()
            w = self._writer
            with self.querying_context(tds_base.PacketType.TRANS):
                self._start_query()
                w.pack(
                    self._begin_tran_struct_72,
                    5,  # TM_BEGIN_XACT
                    isolation_level,
                    0,  # new transaction name
                    )
        else:
            self.submit_plain_query("BEGIN TRANSACTION")
            self.conn.tds72_transaction = 1

    _commit_rollback_tran_struct72_hdr = struct.Struct('<HBB')
    _continue_tran_struct72 = struct.Struct('<BB')

    def rollback(self, cont, isolation_level=0):
        logger.info('Sending ROLLBACK TRAN')
        self.submit_rollback(cont, isolation_level=isolation_level)
        prev_timeout = self._tds.sock.gettimeout()
        self._tds.sock.settimeout(None)
        try:
            self.process_simple_request()
        finally:
            self._tds.sock.settimeout(prev_timeout)

    def submit_rollback(self, cont, isolation_level=0):
        if tds_base.IS_TDS72_PLUS(self):
            self.messages = []
            self.cancel_if_pending()
            w = self._writer
            with self.querying_context(tds_base.PacketType.TRANS):
                self._start_query()
                flags = 0
                if cont:
                    flags |= 1
                w.pack(
                    self._commit_rollback_tran_struct72_hdr,
                    8,  # TM_ROLLBACK_XACT
                    0,  # transaction name
                    flags,
                    )
                if cont:
                    w.pack(
                        self._continue_tran_struct72,
                        isolation_level,
                        0,  # new transaction name
                        )
        else:
            self.submit_plain_query(
                "IF @@TRANCOUNT > 0 ROLLBACK BEGIN TRANSACTION" if cont else "IF @@TRANCOUNT > 0 ROLLBACK")
            self.conn.tds72_transaction = 1 if cont else 0

    def commit(self, cont, isolation_level=0):
        logger.info('Sending COMMIT TRAN')
        self.submit_commit(cont, isolation_level=isolation_level)
        prev_timeout = self._tds.sock.gettimeout()
        self._tds.sock.settimeout(None)
        try:
            self.process_simple_request()
        finally:
            self._tds.sock.settimeout(prev_timeout)

    def submit_commit(self, cont, isolation_level=0):
        if tds_base.IS_TDS72_PLUS(self):
            self.messages = []
            self.cancel_if_pending()
            w = self._writer
            with self.querying_context(tds_base.PacketType.TRANS):
                self._start_query()
                flags = 0
                if cont:
                    flags |= 1
                w.pack(
                    self._commit_rollback_tran_struct72_hdr,
                    7,  # TM_COMMIT_XACT
                    0,  # transaction name
                    flags,
                    )
                if cont:
                    w.pack(
                        self._continue_tran_struct72,
                        isolation_level,
                        0,  # new transaction name
                        )
        else:
            self.submit_plain_query(
                "IF @@TRANCOUNT > 0 COMMIT BEGIN TRANSACTION" if cont else "IF @@TRANCOUNT > 0 COMMIT")
            self.conn.tds72_transaction = 1 if cont else 0

    _tds72_query_start = struct.Struct('<IIHQI')

    def _start_query(self):
        w = self._writer
        w.pack(_TdsSession._tds72_query_start,
               0x16,  # total length
               0x12,  # length
               2,  # type
               self.conn.tds72_transaction,
               1,  # request count
               )

    def send_prelogin(self, login):
        from . import intversion
        # https://msdn.microsoft.com/en-us/library/dd357559.aspx
        instance_name = login.instance_name or 'MSSQLServer'
        instance_name = instance_name.encode('ascii')
        if len(instance_name) > 65490:
            raise ValueError('Instance name is too long')
        if tds_base.IS_TDS72_PLUS(self):
            start_pos = 26
            buf = struct.pack(
                b'>BHHBHHBHHBHHBHHB',
                # netlib version
                PreLoginToken.VERSION, start_pos, 6,
                # encryption
                PreLoginToken.ENCRYPTION, start_pos + 6, 1,
                # instance
                PreLoginToken.INSTOPT, start_pos + 6 + 1, len(instance_name) + 1,
                # thread id
                PreLoginToken.THREADID, start_pos + 6 + 1 + len(instance_name) + 1, 4,
                # MARS enabled
                PreLoginToken.MARS, start_pos + 6 + 1 + len(instance_name) + 1 + 4, 1,
                # end
                PreLoginToken.TERMINATOR
                )
        else:
            start_pos = 21
            buf = struct.pack(
                b'>BHHBHHBHHBHHB',
                # netlib version
                PreLoginToken.VERSION, start_pos, 6,
                # encryption
                PreLoginToken.ENCRYPTION, start_pos + 6, 1,
                # instance
                PreLoginToken.INSTOPT, start_pos + 6 + 1, len(instance_name) + 1,
                # thread id
                PreLoginToken.THREADID, start_pos + 6 + 1 + len(instance_name) + 1, 4,
                # end
                PreLoginToken.TERMINATOR
                )
        assert start_pos == len(buf)
        w = self._writer
        w.begin_packet(tds_base.PacketType.PRELOGIN)
        w.write(buf)
        w.put_uint_be(intversion)
        w.put_usmallint_be(0)  # build number
        # encryption flag
        w.put_byte(login.enc_flag)
        w.write(instance_name)
        w.put_byte(0)  # zero terminate instance_name
        w.put_int(0)  # TODO: change this to thread id
        attribs = {
            'lib_ver': '%x' % intversion,
            'enc_flag': '%x' % login.enc_flag,
            'inst_name': instance_name,
        }
        if tds_base.IS_TDS72_PLUS(self):
            # MARS (1 enabled)
            w.put_byte(1 if login.use_mars else 0)
            attribs['mars'] = login.use_mars
        logger.info('Sending PRELOGIN %s', ' '.join('%s=%s' % (n, v) for n, v in attribs.items()))

        w.flush()

    def process_prelogin(self, login):
        # https://msdn.microsoft.com/en-us/library/dd357559.aspx
        p = self._reader.read_whole_packet()
        size = len(p)
        if size <= 0 or self._reader.packet_type != tds_base.PacketType.REPLY:
            self.bad_stream('Invalid packet type: {0}, expected PRELOGIN(4)'.format(self._reader.packet_type))
        self.parse_prelogin(octets=p, login=login)

    def parse_prelogin(self, octets, login):
        # https://msdn.microsoft.com/en-us/library/dd357559.aspx
        size = len(octets)
        p = octets
        # default 2, no certificate, no encryptption
        crypt_flag = 2
        i = 0
        byte_struct = struct.Struct('B')
        off_len_struct = struct.Struct('>HH')
        prod_version_struct = struct.Struct('>LH')
        while True:
            if i >= size:
                self.bad_stream('Invalid size of PRELOGIN structure')
            type_id, = byte_struct.unpack_from(p, i)
            if type_id == PreLoginToken.TERMINATOR:
                break
            if i + 4 > size:
                self.bad_stream('Invalid size of PRELOGIN structure')
            off, l = off_len_struct.unpack_from(p, i + 1)
            if off > size or off + l > size:
                self.bad_stream('Invalid offset in PRELOGIN structure')
            if type_id == PreLoginToken.VERSION:
                self.conn.server_library_version = prod_version_struct.unpack_from(p, off)
            elif type_id == PreLoginToken.ENCRYPTION and l >= 1:
                crypt_flag, = byte_struct.unpack_from(p, off)
            elif type_id == PreLoginToken.MARS:
                self.conn._mars_enabled = bool(byte_struct.unpack_from(p, off)[0])
            elif type_id == PreLoginToken.INSTOPT:
                # ignore instance name mismatch
                pass
            i += 5
        logger.info("Got PRELOGIN response crypt=%x mars=%d",
                    crypt_flag, self.conn._mars_enabled)
        # if server do not has certificate do normal login
        login.server_enc_flag = crypt_flag
        if crypt_flag == PreLoginEnc.ENCRYPT_OFF:
            if login.enc_flag == PreLoginEnc.ENCRYPT_ON:
                raise self.bad_stream('Server returned unexpected ENCRYPT_ON value')
            else:
                # encrypt login packet only
                tls.establish_channel(self)
        elif crypt_flag == PreLoginEnc.ENCRYPT_ON:
            # encrypt entire connection
            tls.establish_channel(self)
        elif crypt_flag == PreLoginEnc.ENCRYPT_REQ:
            if login.enc_flag == PreLoginEnc.ENCRYPT_NOT_SUP:
                # connection terminated by server and client
                raise tds_base.Error('Client does not have encryption enabled but it is required by server, '
                                     'enable encryption and try connecting again')
            else:
                # encrypt entire connection
                tls.establish_channel(self)
        elif crypt_flag == PreLoginEnc.ENCRYPT_NOT_SUP:
            if login.enc_flag == PreLoginEnc.ENCRYPT_ON:
                # connection terminated by server and client
                raise tds_base.Error('You requested encryption but it is not supported by server')
            # do not encrypt anything
        else:
            self.bad_stream('Unexpected value of enc_flag returned by server: {}'.format(crypt_flag))

    def tds7_send_login(self, login):
        # https://msdn.microsoft.com/en-us/library/dd304019.aspx
        option_flag2 = login.option_flag2
        user_name = login.user_name
        if len(user_name) > 128:
            raise ValueError('User name should be no longer that 128 characters')
        if len(login.password) > 128:
            raise ValueError('Password should be not longer than 128 characters')
        if len(login.change_password) > 128:
            raise ValueError('Password should be not longer than 128 characters')
        if len(login.client_host_name) > 128:
            raise ValueError('Host name should be not longer than 128 characters')
        if len(login.app_name) > 128:
            raise ValueError('App name should be not longer than 128 characters')
        if len(login.server_name) > 128:
            raise ValueError('Server name should be not longer than 128 characters')
        if len(login.database) > 128:
            raise ValueError('Database name should be not longer than 128 characters')
        if len(login.language) > 128:
            raise ValueError('Language should be not longer than 128 characters')
        if len(login.attach_db_file) > 260:
            raise ValueError('File path should be not longer than 260 characters')
        w = self._writer
        w.begin_packet(tds_base.PacketType.LOGIN)
        self.authentication = None
        current_pos = 86 + 8 if tds_base.IS_TDS72_PLUS(self) else 86
        client_host_name = login.client_host_name
        login.client_host_name = client_host_name
        packet_size = current_pos + (len(client_host_name) + len(login.app_name) + len(login.server_name) +
                                     len(login.library) + len(login.language) + len(login.database)) * 2
        if login.auth:
            self.authentication = login.auth
            auth_packet = login.auth.create_packet()
            packet_size += len(auth_packet)
        else:
            auth_packet = ''
            packet_size += (len(user_name) + len(login.password)) * 2
        w.put_int(packet_size)
        w.put_uint(login.tds_version)
        w.put_int(w.bufsize)
        from . import intversion
        w.put_uint(intversion)
        w.put_int(login.pid)
        w.put_uint(0)  # connection id
        option_flag1 = tds_base.TDS_SET_LANG_ON | tds_base.TDS_USE_DB_NOTIFY | tds_base.TDS_INIT_DB_FATAL
        if not login.bulk_copy:
            option_flag1 |= tds_base.TDS_DUMPLOAD_OFF
        w.put_byte(option_flag1)
        if self.authentication:
            option_flag2 |= tds_base.TDS_INTEGRATED_SECURITY_ON
        w.put_byte(option_flag2)
        type_flags = 0
        if login.readonly:
            type_flags |= (2 << 5)
        w.put_byte(type_flags)
        option_flag3 = tds_base.TDS_UNKNOWN_COLLATION_HANDLING
        w.put_byte(option_flag3 if tds_base.IS_TDS73_PLUS(self) else 0)
        mins_fix = int(login.client_tz.utcoffset(datetime.datetime.now()).total_seconds()) // 60
        logger.info('Sending LOGIN tds_ver=%x bufsz=%d pid=%d opt1=%x opt2=%x opt3=%x cli_tz=%d cli_lcid=%s '
                    'cli_host=%s lang=%s db=%s',
                    login.tds_version, w.bufsize, login.pid, option_flag1, option_flag2, option_flag3, mins_fix,
                    login.client_lcid, client_host_name, login.language, login.database)
        w.put_int(mins_fix)
        w.put_int(login.client_lcid)
        w.put_smallint(current_pos)
        w.put_smallint(len(client_host_name))
        current_pos += len(client_host_name) * 2
        if self.authentication:
            w.put_smallint(0)
            w.put_smallint(0)
            w.put_smallint(0)
            w.put_smallint(0)
        else:
            w.put_smallint(current_pos)
            w.put_smallint(len(user_name))
            current_pos += len(user_name) * 2
            w.put_smallint(current_pos)
            w.put_smallint(len(login.password))
            current_pos += len(login.password) * 2
        w.put_smallint(current_pos)
        w.put_smallint(len(login.app_name))
        current_pos += len(login.app_name) * 2
        # server name
        w.put_smallint(current_pos)
        w.put_smallint(len(login.server_name))
        current_pos += len(login.server_name) * 2
        # reserved
        w.put_smallint(0)
        w.put_smallint(0)
        # library name
        w.put_smallint(current_pos)
        w.put_smallint(len(login.library))
        current_pos += len(login.library) * 2
        # language
        w.put_smallint(current_pos)
        w.put_smallint(len(login.language))
        current_pos += len(login.language) * 2
        # database name
        w.put_smallint(current_pos)
        w.put_smallint(len(login.database))
        current_pos += len(login.database) * 2
        # ClientID
        client_id = struct.pack('>Q', login.client_id)[2:]
        w.write(client_id)
        # authentication
        w.put_smallint(current_pos)
        w.put_smallint(len(auth_packet))
        current_pos += len(auth_packet)
        # db file
        w.put_smallint(current_pos)
        w.put_smallint(len(login.attach_db_file))
        current_pos += len(login.attach_db_file) * 2
        if tds_base.IS_TDS72_PLUS(self):
            # new password
            w.put_smallint(current_pos)
            w.put_smallint(len(login.change_password))
            # sspi long
            w.put_int(0)
        w.write_ucs2(client_host_name)
        if not self.authentication:
            w.write_ucs2(user_name)
            w.write(tds7_crypt_pass(login.password))
        w.write_ucs2(login.app_name)
        w.write_ucs2(login.server_name)
        w.write_ucs2(login.library)
        w.write_ucs2(login.language)
        w.write_ucs2(login.database)
        if self.authentication:
            w.write(auth_packet)
        w.write_ucs2(login.attach_db_file)
        w.write_ucs2(login.change_password)
        w.flush()

    _SERVER_TO_CLIENT_MAPPING = {
        0x07000000: tds_base.TDS70,
        0x07010000: tds_base.TDS71,
        0x71000001: tds_base.TDS71rev1,
        tds_base.TDS72: tds_base.TDS72,
        tds_base.TDS73A: tds_base.TDS73A,
        tds_base.TDS73B: tds_base.TDS73B,
        tds_base.TDS74: tds_base.TDS74,
        }

    def process_login_tokens(self):
        r = self._reader
        succeed = False
        while True:
            marker = r.get_byte()
            if marker == tds_base.TDS_LOGINACK_TOKEN:
                # https://msdn.microsoft.com/en-us/library/dd340651.aspx
                succeed = True
                size = r.get_smallint()
                r.get_byte()  # interface
                version = r.get_uint_be()
                self.conn.tds_version = self._SERVER_TO_CLIENT_MAPPING.get(version, version)
                if not tds_base.IS_TDS7_PLUS(self):
                    self.bad_stream('Only TDS 7.0 and higher are supported')
                # get server product name
                # ignore product name length, some servers seem to set it incorrectly
                r.get_byte()
                size -= 10
                self.conn.product_name = r.read_ucs2(size // 2)
                product_version = r.get_uint_be()
                logger.info('Got LOGINACK tds_ver=%x srv_name=%s srv_ver=%x',
                            self.conn.tds_version, self.conn.product_name, product_version)
                # MSSQL 6.5 and 7.0 seem to return strange values for this
                # using TDS 4.2, something like 5F 06 32 FF for 6.50
                self.conn.product_version = product_version
                if self.authentication:
                    self.authentication.close()
                    self.authentication = None
            else:
                self.process_token(marker)
                if marker == tds_base.TDS_DONE_TOKEN:
                    break
        return succeed

    def process_returnstatus(self):
        self.log_response_message('got RETURNSTATUS message')
        self.ret_status = self._reader.get_int()
        self.has_status = True

    def process_token(self, marker):
        handler = _token_map.get(marker)
        if not handler:
            self.bad_stream('Invalid TDS marker: {0}({0:x})'.format(marker))
        return handler(self)

    def get_token_id(self):
        self.set_state(tds_base.TDS_READING)
        try:
            marker = self._reader.get_byte()
        except tds_base.TimeoutError:
            self.set_state(tds_base.TDS_PENDING)
            raise
        except:
            self._tds.close()
            raise
        return marker

    def process_simple_request(self):
        while True:
            marker = self.get_token_id()
            if marker in (tds_base.TDS_DONE_TOKEN, tds_base.TDS_DONEPROC_TOKEN, tds_base.TDS_DONEINPROC_TOKEN):
                self.process_end(marker)
                if not self.done_flags & tds_base.TDS_DONE_MORE_RESULTS:
                    return
            else:
                self.process_token(marker)

    def next_set(self):
        while self.more_rows:
            self.next_row()
        if self.state == tds_base.TDS_IDLE:
            return False
        if self.find_result_or_done():
            return True

    def fetchone(self):
        if self.res_info is None:
            raise tds_base.ProgrammingError("Previous statement didn't produce any results")

        if self.skipped_to_status:
            raise tds_base.ProgrammingError("Unable to fetch any rows after accessing return_status")

        if not self.next_row():
            return None

        return self.row

    def next_row(self):
        if not self.more_rows:
            return False
        while True:
            marker = self.get_token_id()
            if marker in (tds_base.TDS_ROW_TOKEN, tds_base.TDS_NBC_ROW_TOKEN):
                self.process_token(marker)
                return True
            elif marker in (tds_base.TDS_DONE_TOKEN, tds_base.TDS_DONEPROC_TOKEN, tds_base.TDS_DONEINPROC_TOKEN):
                self.process_end(marker)
                return False
            else:
                self.process_token(marker)

    def find_result_or_done(self):
        self.done_flags = 0
        while True:
            marker = self.get_token_id()
            if marker == tds_base.TDS7_RESULT_TOKEN:
                self.process_token(marker)
                return True
            elif marker in (tds_base.TDS_DONE_TOKEN, tds_base.TDS_DONEPROC_TOKEN, tds_base.TDS_DONEINPROC_TOKEN):
                self.process_end(marker)
                if self.done_flags & tds_base.TDS_DONE_MORE_RESULTS:
                    if self.done_flags & tds_base.TDS_DONE_COUNT:
                        return True
                else:
                    return False
            else:
                self.process_token(marker)

    def process_rpc(self):
        self.done_flags = 0
        self.return_value_index = 0
        while True:
            marker = self.get_token_id()
            if marker == tds_base.TDS7_RESULT_TOKEN:
                self.process_token(marker)
                return True
            elif marker in (tds_base.TDS_DONE_TOKEN, tds_base.TDS_DONEPROC_TOKEN):
                self.process_end(marker)
                if self.done_flags & tds_base.TDS_DONE_MORE_RESULTS and not self.done_flags & tds_base.TDS_DONE_COUNT:
                    # skip results that don't event have rowcount
                    continue
                return False
            else:
                self.process_token(marker)

    def complete_rpc(self):
        # go through all result sets
        while self.next_set():
            pass

    def find_return_status(self):
        self.skipped_to_status = True
        while True:
            marker = self.get_token_id()
            self.process_token(marker)
            if marker == tds_base.TDS_RETURNSTATUS_TOKEN:
                return


_token_map = {
    tds_base.TDS_AUTH_TOKEN: _TdsSession.process_auth,
    tds_base.TDS_ENVCHANGE_TOKEN: _TdsSession.process_env_chg,
    tds_base.TDS_DONE_TOKEN: lambda self: self.process_end(tds_base.TDS_DONE_TOKEN),
    tds_base.TDS_DONEPROC_TOKEN: lambda self: self.process_end(tds_base.TDS_DONEPROC_TOKEN),
    tds_base.TDS_DONEINPROC_TOKEN: lambda self: self.process_end(tds_base.TDS_DONEINPROC_TOKEN),
    tds_base.TDS_ERROR_TOKEN: lambda self: self.process_msg(tds_base.TDS_ERROR_TOKEN),
    tds_base.TDS_INFO_TOKEN: lambda self: self.process_msg(tds_base.TDS_INFO_TOKEN),
    tds_base.TDS_CAPABILITY_TOKEN: lambda self: self.process_msg(tds_base.TDS_CAPABILITY_TOKEN),
    tds_base.TDS_PARAM_TOKEN: lambda self: self.process_param(),
    tds_base.TDS7_RESULT_TOKEN: lambda self: self.tds7_process_result(),
    tds_base.TDS_ROW_TOKEN: lambda self: self.process_row(),
    tds_base.TDS_NBC_ROW_TOKEN: lambda self: self.process_nbcrow(),
    tds_base.TDS_ORDERBY_TOKEN: lambda self: self.process_orderby(),
    tds_base.TDS_RETURNSTATUS_TOKEN: lambda self: self.process_returnstatus(),
    }


# this class represents root TDS connection
# if MARS is used it can have multiple sessions represented by _TdsSession class
# if MARS is not used it would have single _TdsSession instance
class _TdsSocket(object):
    def __init__(self, use_tz=None):
        self._is_connected = False
        self.env = _TdsEnv()
        self.collation = None
        self.tds72_transaction = 0
        self._mars_enabled = False
        self.sock = None
        self.bufsize = 4096
        self.tds_version = tds_base.TDS74
        self.use_tz = use_tz
        self.type_factory = tds_types.SerializerFactory(self.tds_version)
        self.type_inferrer = None
        self.query_timeout = 0
        self._smp_manager = None
        self._main_session = None
        self._login = None
        self.route = None

    def __repr__(self):
        fmt = "<_TdsSocket tran={} mars={} tds_version={} use_tz={}>"
        return fmt.format(self.tds72_transaction, self._mars_enabled,
                          self.tds_version, self.use_tz)

    def login(self, login, sock, tzinfo_factory):
        self._login = login
        self.bufsize = login.blocksize
        self.query_timeout = login.query_timeout
        self._main_session = _TdsSession(self, sock, tzinfo_factory)
        self.sock = sock
        self.tds_version = login.tds_version
        login.server_enc_flag = PreLoginEnc.ENCRYPT_NOT_SUP
        if tds_base.IS_TDS71_PLUS(self):
            self._main_session.send_prelogin(login)
            self._main_session.process_prelogin(login)
        self._main_session.tds7_send_login(login)
        if login.server_enc_flag == PreLoginEnc.ENCRYPT_OFF:
            tls.revert_to_clear(self._main_session)
        if not self._main_session.process_login_tokens():
            self._main_session.raise_db_exception()
        if self.route is not None:
            return self.route

        # update block size if server returned different one
        if self._main_session._writer.bufsize != self._main_session._reader.get_block_size():
            self._main_session._reader.set_block_size(self._main_session._writer.bufsize)

        self.type_factory = tds_types.SerializerFactory(self.tds_version)
        self.type_inferrer = tds_types.TdsTypeInferrer(
            type_factory=self.type_factory,
            collation=self.collation,
            bytes_to_unicode=self._login.bytes_to_unicode,
            allow_tz=not self.use_tz
        )
        if self._mars_enabled:
            from .smp import SmpManager
            self._smp_manager = SmpManager(self.sock)
            self._main_session = _TdsSession(
                self,
                self._smp_manager.create_session(),
                tzinfo_factory)
        self._is_connected = True
        q = []
        if login.database and self.env.database != login.database:
            q.append('use ' + tds_base.tds_quote_id(login.database))
        if q:
            self._main_session.submit_plain_query(''.join(q))
            self._main_session.process_simple_request()
        return None

    @property
    def mars_enabled(self):
        return self._mars_enabled

    @property
    def main_session(self):
        return self._main_session

    def create_session(self, tzinfo_factory):
        return _TdsSession(
            self, self._smp_manager.create_session(),
            tzinfo_factory)

    def is_connected(self):
        return self._is_connected

    def close(self):
        self._is_connected = False
        if self.sock is not None:
            self.sock.close()
        if self._smp_manager:
            self._smp_manager.transport_closed()
        self._main_session.state = tds_base.TDS_DEAD
        if self._main_session.authentication:
            self._main_session.authentication.close()
            self._main_session.authentication = None


class _Results(object):
    def __init__(self):
            self.columns = []
            self.row_count = 0


def _parse_instances(msg):
    name = None
    if len(msg) > 3 and tds_base.my_ord(msg[0]) == 5:
        tokens = msg[3:].decode('ascii').split(';')
        results = {}
        instdict = {}
        got_name = False
        for token in tokens:
            if got_name:
                instdict[name] = token
                got_name = False
            else:
                name = token
                if not name:
                    if not instdict:
                        break
                    results[instdict['InstanceName'].upper()] = instdict
                    instdict = {}
                    continue
                got_name = True
        return results


#
# Get port of all instances
# @return default port number or 0 if error
# @remark experimental, cf. MC-SQLR.pdf.
#
def tds7_get_instances(ip_addr, timeout=5):
    s = socket.socket(type=socket.SOCK_DGRAM)
    s.settimeout(timeout)
    try:
        # send the request
        s.sendto(b'\x03', (ip_addr, 1434))
        msg = s.recv(16 * 1024 - 1)
        # got data, read and parse
        return _parse_instances(msg)
    finally:
        s.close()



import datetime
from pytds import tz


def test_tz():
    assert tz.FixedOffsetTimezone(0, 'UTC').tzname(None) == 'UTC'
    lz = tz.LocalTimezone()
    jan_1 = datetime.datetime(2010, 1, 1, 0, 0)
    july_1 = datetime.datetime(2010, 7, 1, 0, 0)
    assert isinstance(lz.tzname(jan_1), str)
    lz.dst(jan_1)
    lz.dst(july_1)
    lz.utcoffset(jan_1)
    lz.utcoffset(july_1)



# coding=utf-8
import datetime
from decimal import Decimal, Context
import uuid

import pytest
import six

import pytds

from fixtures import *


@pytest.mark.parametrize('sql_type', [
    "tinyint",
    "smallint",
    "int",
    "bigint",
    "real",
    "float",
    "smallmoney",
    "money",
    "decimal",
    "varbinary(15)",
    "binary(15)",
    "nvarchar(15)",
    "nchar(15)",
    "varchar(15)",
    "char(15)",
    "bit",
    "smalldatetime",
    "date",
    "time",
    "datetime",
    "datetime2",
    "datetimeoffset",
    "uniqueidentifier",
    "sql_variant",
])
def test_null_parameter(cursor, sql_type):
    cursor.execute("set nocount on; declare @x {} = %s; select @x".format(sql_type), (None,))
    val, = cursor.fetchone()
    assert val is None


def test_reading_values(cursor):
    cur = cursor
    with pytest.raises(pytds.ProgrammingError):
        cur.execute(u'select ')
    assert 'abc' == cur.execute_scalar("select cast('abc' as varchar(max)) as fieldname")
    assert 'abc' == cur.execute_scalar("select cast('abc' as nvarchar(max)) as fieldname")
    assert b'abc' == cur.execute_scalar("select cast('abc' as varbinary(max)) as fieldname")
    #assert 12 == cur.execute_scalar('select cast(12 as bigint) as fieldname')
    assert 12 == cur.execute_scalar('select cast(12 as smallint) as fieldname')
    assert -12 == cur.execute_scalar('select -12 as fieldname')
    assert 12 == cur.execute_scalar('select cast(12 as tinyint) as fieldname')
    assert True == cur.execute_scalar('select cast(1 as bit) as fieldname')
    assert 5.1 == cur.execute_scalar('select cast(5.1 as float) as fieldname')
    cur.execute("select 'test', 20")
    assert ('test', 20) == cur.fetchone()
    assert 'test' == cur.execute_scalar("select 'test' as fieldname")
    assert 'test' == cur.execute_scalar("select N'test' as fieldname")
    assert 'test' == cur.execute_scalar("select cast(N'test' as ntext) as fieldname")
    assert 'test' == cur.execute_scalar("select cast(N'test' as text) as fieldname")
    assert 'test ' == cur.execute_scalar("select cast(N'test' as char(5)) as fieldname")
    assert 'test ' == cur.execute_scalar("select cast(N'test' as nchar(5)) as fieldname")
    assert b'test' == cur.execute_scalar("select cast('test' as varbinary(4)) as fieldname")
    assert b'test' == cur.execute_scalar("select cast('test' as image) as fieldname")
    assert None == cur.execute_scalar("select cast(NULL as image) as fieldname")
    assert None == cur.execute_scalar("select cast(NULL as varbinary(10)) as fieldname")
    assert None == cur.execute_scalar("select cast(NULL as ntext) as fieldname")
    assert None == cur.execute_scalar("select cast(NULL as nvarchar(max)) as fieldname")
    assert None == cur.execute_scalar("select cast(NULL as xml)")
    assert None is cur.execute_scalar("select cast(NULL as varchar(max)) as fieldname")
    assert None == cur.execute_scalar("select cast(NULL as nvarchar(10)) as fieldname")
    assert None == cur.execute_scalar("select cast(NULL as varchar(10)) as fieldname")
    assert None == cur.execute_scalar("select cast(NULL as nchar(10)) as fieldname")
    assert None == cur.execute_scalar("select cast(NULL as char(10)) as fieldname")
    assert None == cur.execute_scalar("select cast(NULL as char(10)) as fieldname")
    assert 5 == cur.execute_scalar('select 5 as fieldname')
    with pytest.raises(pytds.ProgrammingError) as ex:
        cur.execute_scalar('create table exec_scalar_empty(f int)')
    # message does not have to be exact match
    assert "Previous statement didn't produce any results" in str(ex.value)


def test_money(cursor):
    cur = cursor
    assert Decimal('0') == cur.execute_scalar("select cast('0' as money) as fieldname")
    assert Decimal('1') == cur.execute_scalar("select cast('1' as money) as fieldname")
    assert Decimal('1.5555') == cur.execute_scalar("select cast('1.5555' as money) as fieldname")
    assert Decimal('1234567.5555') == cur.execute_scalar("select cast('1234567.5555' as money) as fieldname")
    assert Decimal('-1234567.5555') == cur.execute_scalar("select cast('-1234567.5555' as money) as fieldname")
    assert Decimal('12345.55') == cur.execute_scalar("select cast('12345.55' as smallmoney) as fieldname")


def test_strs(cursor):
    cur = cursor
    assert isinstance(cur.execute_scalar("select 'test'"), six.text_type)


@pytest.mark.parametrize('val', [u'hello',
                                 u'x' * 5000,
                                 'x' * 9000,
                                 123,
                                 -123,
                                 123.12,
                                 -123.12,
                                 10 ** 20,
                                 10 ** 38 - 1,
                                 -10 ** 38 + 1,
                                 datetime.datetime(2011, 2, 3, 10, 11, 12, 3000),
                                 Decimal('1234.567'),
                                 Decimal('1234000'),
                                 Decimal('9' * 38),
                                 Decimal('0.' + '9' * 38),
                                 Decimal('-' + ('9' * 38), Context(prec=38)),
                                 Decimal('1E10'),
                                 Decimal('1E-10'),
                                 Decimal('0.{0}1'.format('0' * 37)),
                                 None,
                                 'hello',
                                 '',
                                 pytds.Binary(b''),
                                 pytds.Binary(b'\x00\x01\x02'),
                                 pytds.Binary(b'x' * 9000),
                                 2 ** 63 - 1,
                                 False,
                                 True,
                                 uuid.uuid4(),
                                 u'Iñtërnâtiônàlizætiøn1',
                                 u'\U0001d6fc',
                                 ])
def test_select_values(cursor, val):
    cursor.execute('select %s', (val,))
    assert cursor.fetchone() == (val,)
    assert cursor.fetchone() is None


uuid_val = uuid.uuid4()


@pytest.mark.parametrize('result,sql', [
    (None, "cast(NULL as varchar)"),
    ('test', "cast('test' as varchar)"),
    ('test ', "cast('test' as char(5))"),
    ('test', "cast(N'test' as nvarchar)"),
    ('test ', "cast(N'test' as nchar(5))"),
    (Decimal('100.55555'), "cast(100.55555 as decimal(8,5))"),
    (Decimal('100.55555'), "cast(100.55555 as numeric(8,5))"),
    (b'test', "cast('test' as varbinary)"),
    (b'test\x00', "cast('test' as binary(5))"),
    (datetime.datetime(2011, 2, 3, 10, 11, 12, 3000), "cast('2011-02-03T10:11:12.003' as datetime)"),
    (datetime.datetime(2011, 2, 3, 10, 11, 0), "cast('2011-02-03T10:11:00' as smalldatetime)"),
    (uuid_val, "cast('{0}' as uniqueidentifier)".format(uuid_val)),
    (True, "cast(1 as bit)"),
    (128, "cast(128 as tinyint)"),
    (255, "cast(255 as tinyint)"),
    (-32000, "cast(-32000 as smallint)"),
    (2000000000, "cast(2000000000 as int)"),
    (2000000000000, "cast(2000000000000 as bigint)"),
    (0.12345, "cast(0.12345 as float)"),
    (0.25, "cast(0.25 as real)"),
    (Decimal('922337203685477.5807'), "cast('922,337,203,685,477.5807' as money)"),
    (Decimal('-214748.3648'), "cast('- 214,748.3648' as smallmoney)"),
])
def test_sql_variant_round_trip(cursor, result, sql):
    if not pytds.tds_base.IS_TDS71_PLUS(cursor.connection):
        pytest.skip('Requires TDS7.1+')
    cursor.execute("select cast({0} as sql_variant)".format(sql))
    val, = cursor.fetchone()
    assert result == val


def test_collations(cursor, collation_set):
    coll_name_set = collation_set

    tests = [
        ('Привет', 'Cyrillic_General_BIN'),
        ('Привет', 'Cyrillic_General_BIN2'),
        ('สวัสดี', 'Thai_CI_AI'),
        ('你好', 'Chinese_PRC_CI_AI'),
        ('こんにちは', 'Japanese_CI_AI'),
        ('안녕하세요.', 'Korean_90_CI_AI'),
        ('你好', 'Chinese_Hong_Kong_Stroke_90_CI_AI'),
        ('cześć', 'Polish_CI_AI'),
        ('Bonjour', 'French_CI_AI'),
        ('Γεια σας', 'Greek_CI_AI'),
        ('Merhaba', 'Turkish_CI_AI'),
        ('שלום', 'Hebrew_CI_AI'),
        ('مرحبا', 'Arabic_CI_AI'),
        ('Sveiki', 'Lithuanian_CI_AI'),
        ('chào', 'Vietnamese_CI_AI'),
        ('ÄÅÆ', 'SQL_Latin1_General_CP437_BIN'),
        ('ÁÂÀÃ', 'SQL_Latin1_General_CP850_BIN'),
        ('ŠşĂ', 'SQL_Slovak_CP1250_CS_AS_KI_WI'),
        ('ÁÂÀÃ', 'SQL_Latin1_General_1251_BIN'),
        ('ÁÂÀÃ', 'SQL_Latin1_General_Cp1_CS_AS_KI_WI'),
        ('ÁÂÀÃ', 'SQL_Latin1_General_1253_BIN'),
        ('ÁÂÀÃ', 'SQL_Latin1_General_1254_BIN'),
        ('ÁÂÀÃ', 'SQL_Latin1_General_1255_BIN'),
        ('ÁÂÀÃ', 'SQL_Latin1_General_1256_BIN'),
        ('ÁÂÀÃ', 'SQL_Latin1_General_1257_BIN'),
        ('ÁÂÀÃ', 'Latin1_General_100_BIN'),
    ]
    for s, coll in tests:
        if coll not in coll_name_set:
            logger.info('Skipping {}, not supported by current server'.format(coll))
            continue
        assert cursor.execute_scalar("select cast(N'{}' collate {} as varchar(100))".format(s, coll)) == s


def skip_if_new_date_not_supported(conn):
    if not pytds.tds_base.IS_TDS73_PLUS(conn):
        pytest.skip('Test requires new date types support, SQL 2008 or newer is required')


def test_date(cursor):
    skip_if_new_date_not_supported(cursor.connection)
    date = pytds.Date(2012, 10, 6)
    cursor.execute('select %s', (date, ))
    assert cursor.fetchall() == [(date,)]


def test_time(cursor):
    skip_if_new_date_not_supported(cursor.connection)
    time = pytds.Time(8, 7, 4, 123000)
    cursor.execute('select %s', (time, ))
    assert cursor.fetchall() == [(time,)]


def test_datetime(cursor):
    time = pytds.Timestamp(2013, 7, 9, 8, 7, 4, 123000)
    cursor.execute('select %s', (time, ))
    assert cursor.fetchall() == [(time,)]
